"""Dealing with file infos of the Notion API."""

from __future__ import annotations

import io
import time
from abc import ABC, abstractmethod
from typing import BinaryIO, cast
from urllib.parse import urlparse
from uuid import UUID

import filetype
import pendulum as pnd
from typing_extensions import Self, TypeVar
from url_normalize import url_normalize

from ultimate_notion.core import Wrapper, get_active_session, get_repr
from ultimate_notion.errors import InvalidAPIUsageError
from ultimate_notion.obj_api import objects as objs
from ultimate_notion.obj_api.core import raise_unset
from ultimate_notion.obj_api.enums import FileUploadStatus
from ultimate_notion.rich_text import Text, html_img

NOTION_HOSTED_DOMAIN = 'secure.notion-static.com'
MAX_FILE_SIZE = 20_000_000
"""Maximum file size for single part upload. It's 5MB only for the free plan"""


# ToDo: Use new syntax when requires-python >= 3.12
FO_co = TypeVar('FO_co', bound=objs.FileObject, default=objs.FileObject, covariant=True)


class AnyFile(Wrapper[FO_co], ABC, wraps=objs.FileObject):
    """Information about a web resource e.g. for the files property."""

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, str | AnyFile):
            return NotImplemented
        return str(self) == str(other)

    def __hash__(self) -> int:
        return hash(str(self))

    @abstractmethod
    def __str__(self) -> str: ...

    @property
    def name(self) -> str | None:
        """Return the name of the file."""
        return self.obj_ref.name

    @name.setter
    def name(self, name: str | None) -> None:
        self.obj_ref.name = name

    @property
    def caption(self) -> Text | None:
        """Return the caption of the file."""
        if (caption := self.obj_ref.caption) is None:
            return None
        return Text.wrap_obj_ref(caption)

    @caption.setter
    def caption(self, caption: str | None) -> None:
        self.obj_ref.caption = Text(caption).obj_ref if caption is not None else []


class NotionFile(AnyFile[objs.HostedFile], wraps=objs.HostedFile):
    """Information about a file that is hosted by Notion."""

    def __init__(self, *, url: str, name: str | None = None, caption: str | None = None) -> None:
        caption_obj = Text(caption).obj_ref if caption is not None else None  # [] is not accepted here by the API!
        super().__init__(url=url_normalize(url), name=name, caption=caption_obj)

    def __str__(self) -> str:
        return f'NotionFile({self.url})'

    def __repr__(self) -> str:
        return get_repr(self, desc=f'url={self.url}')

    def _repr_html_(self) -> str:  # noqa: PLW3201
        """Called by JupyterLab automatically to display this file."""
        return html_img(self.url, size=2)

    @property
    def url(self) -> str:
        return self.obj_ref.file.url


class ExternalFile(AnyFile[objs.ExternalFile], wraps=objs.ExternalFile):
    """Information about a file that is hosted externally, i.e. not by Notion."""

    def __init__(self, *, url: str, name: str | None = None, caption: str | None = None) -> None:
        caption_obj = Text(caption).obj_ref if caption is not None else None  # [] is not accepted here by the API!
        super().__init__(url=url_normalize(url), name=name, caption=caption_obj)

    def __str__(self) -> str:
        return f'ExternalFile({self.url})'

    def __repr__(self) -> str:
        return get_repr(self, desc=f'url={self.url}')

    def _repr_html_(self) -> str:  # noqa: PLW3201
        """Called by JupyterLab automatically to display this file."""
        return html_img(self.url, size=2)

    @property
    def url(self) -> str:
        return self.obj_ref.external.url


class UploadedFile(AnyFile[objs.UploadedFile], wraps=objs.UploadedFile):
    """Information about a file that has been uploaded to Notion.

    !!! note

        This class is used to represent files that have been uploaded to Notion.
        After it has been used, e.g. to change a cover or add a file block,
        it will be converted to a `NotionFile` (i.e. `objs.HostedFile`), when
        read again from the API.
    """

    poll_interval: float = 1.0
    obj_file_upload: objs.FileUpload

    def __init__(self) -> None:
        msg = 'Use the corresponding methods of a `Session` to get an uploaded file object.'
        raise InvalidAPIUsageError(msg)

    def __str__(self) -> str:
        return f'UploadedFile({self.id})'

    def __repr__(self) -> str:
        return get_repr(self, desc=f'id={self.id}')

    def _repr_html_(self) -> str:  # noqa: PLW3201
        """Called by JupyterLab automatically to display this file."""
        return html_img(str(self.id), size=2)

    @property
    def id(self) -> UUID:
        """Return the ID of the uploaded file."""
        return raise_unset(self.obj_file_upload.id)

    @classmethod
    def from_file_upload(cls, file_upload: objs.FileUpload) -> Self:
        """Create an UploadedFile instance from a FileUpload object."""
        file_obj = objs.UploadedFile.build(id=file_upload.id)
        self = cls.__new__(cls)
        self.obj_ref = file_obj
        self.obj_file_upload = file_upload
        return self

    @property
    def status(self) -> FileUploadStatus:
        """Return the status of the uploaded file."""
        return self.obj_file_upload.status

    @property
    def file_name(self) -> str | None:
        """Return the file name of the uploaded file."""
        return self.obj_file_upload.filename

    @property
    def expiry_time(self) -> pnd.DateTime | None:
        """Return the expiry time of the uploaded file."""
        if expire_time := self.obj_file_upload.expiry_time:
            return pnd.instance(expire_time)
        return None

    @property
    def content_type(self) -> str | None:
        """Return the content type of the uploaded file."""
        return self.obj_file_upload.content_type

    @property
    def content_length(self) -> int | None:
        """Return the content length of the uploaded file."""
        return self.obj_file_upload.content_length

    @property
    def file_import_result(self) -> objs.FileImportSuccess | objs.FileImportError | None:
        """Return the file import result of the uploaded file."""
        if result := self.obj_file_upload.file_import_result:
            return result
        return None

    def update_status(self) -> Self:
        """Update the uploaded file information."""
        session = get_active_session()
        self.obj_file_upload = session.api.uploads.retrieve(self.id)
        return self

    def wait_until_uploaded(self) -> Self:
        """Wait until the uploaded file is fully processed."""
        while self.status != FileUploadStatus.UPLOADED:
            time.sleep(self.poll_interval)
            self.update_status()
        return self


def is_notion_hosted(url: str) -> bool:
    """Check if the URL is hosted on Notion."""
    return urlparse(url).netloc == NOTION_HOSTED_DOMAIN


def get_file_size(file: BinaryIO) -> int:
    """Get the size of a file in bytes.

    This function preserves the current file position.

    Args:
        file: The binary file object to measure

    Returns:
        The size of the file in bytes
    """
    current_pos = file.tell()
    file.seek(0, io.SEEK_END)
    file_size = file.tell()
    file.seek(current_pos)  # Reset to original position
    return file_size


def get_mime_type(file: BinaryIO) -> str:
    """Detect the MIME type of a file.

    This function preserves the current file position.

    Args:
        file: The binary file object to analyze

    Returns:
        The detected MIME type, or 'application/octet-stream' if unknown
    """
    current_pos = file.tell()
    content_sample = file.read(1024)
    file.seek(current_pos)  # Reset position

    # correct some MIME types to match Notion's expectations
    mime_type_map = {
        'audio/x-wav': 'audio/wav',  # see issue #127
    }

    kind = filetype.guess(content_sample)
    if kind is not None:
        return cast(str, mime_type_map.get(kind.mime, kind.mime))
    else:
        return 'application/octet-stream'


def url(url: str, *, name: str | None = None, caption: str | None = None) -> NotionFile | ExternalFile:
    """Create a NotionFile or ExternalFile based on the URL.

    A name and caption can be provided and will be used as default values, e.g. in a File block.
    """
    if is_notion_hosted(url):
        return NotionFile(url=url, name=name, caption=caption)
    else:
        return ExternalFile(url=url, name=name, caption=caption)



"""Core classes and functions for the Ultimate Notion API."""

from __future__ import annotations

import datetime as dt
import logging
from abc import ABC
from enum import Enum
from typing import TYPE_CHECKING, Any, ClassVar, Final, Generic, Literal, TypeAlias, cast
from uuid import UUID

from notion_client.errors import APIResponseError
from typing_extensions import Self, TypeVar

from ultimate_notion.obj_api import core as obj_core
from ultimate_notion.obj_api import objects as objs
from ultimate_notion.obj_api.core import raise_unset

_logger = logging.getLogger(__name__)

# ToDo: Use new syntax when requires-python >= 3.12
GT_co = TypeVar('GT_co', bound=obj_core.GenericObject, default=obj_core.GenericObject, covariant=True)

if TYPE_CHECKING:
    from pydantic_core import SchemaSerializer

    from ultimate_notion.session import Session
    from ultimate_notion.user import User


class Wrapper(Generic[GT_co], ABC):
    """Convert objects from the obj-based API to the high-level API and vice versa."""

    _obj_ref: GT_co

    _obj_api_map: ClassVar[dict[type[GT_co], type[Wrapper]]] = {}  # type: ignore[misc]

    def __init_subclass__(cls, wraps: type[GT_co], **kwargs: Any):
        super().__init_subclass__(**kwargs)
        cls._obj_api_map[wraps] = cls

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        # Needed for wrap_obj_ref and its call to __new__ to work!
        return super().__new__(cls)

    def __init__(self, *args: Any, **kwargs: Any):
        """Default constructor that also builds `obj_ref`."""
        obj_api_type: type[GT_co] = self._obj_api_map_inv[self.__class__]
        self._obj_ref = obj_api_type.build(*args, **kwargs)

    def __pydantic_serializer__(self) -> SchemaSerializer:  # noqa: PLW3201
        """Return the Pydantic serializers for this object."""
        # This is used only when creating a pydantic model from a schema.
        return self._obj_ref.__pydantic_serializer__

    @property
    def obj_ref(self) -> GT_co:
        """Return the low-level Notion-API object reference.

        This is just the answer of the Notion API as a Pydantic model.
        """
        return self._obj_ref

    @obj_ref.setter
    def obj_ref(self, value: GT_co) -> None:  # type: ignore[misc] # breaking covariance
        """Set the low-level Notion-API object reference."""
        self._obj_ref = value

    @classmethod
    def wrap_obj_ref(cls: type[Self], obj_ref: GT_co, /) -> Self:  # type: ignore[misc] # breaking covariance
        """Wraps low-level `obj_ref` from Notion API into a high-level (hl) object of Ultimate Notion."""
        hl_cls = cls._obj_api_map[type(obj_ref)]
        # To allow for `Block.wrap_obj_ref` to work call a specific 'wrap_obj_ref' if it exists,
        # e.g. `RichText.wrap_obj_ref` we need to break a potential recursion in the MRO.
        if cls.wrap_obj_ref.__func__ is hl_cls.wrap_obj_ref.__func__:  # type: ignore[attr-defined]
            # ToDo: remove type ignore when https://github.com/python/mypy/issues/14123 is fixed
            # break recursion
            hl_obj = hl_cls.__new__(hl_cls)
            hl_obj._obj_ref = obj_ref
            return cast(Self, hl_obj)
        else:
            return cast(Self, hl_cls.wrap_obj_ref(obj_ref))

    @property
    def _obj_api_map_inv(self) -> dict[type[Wrapper], type[GT_co]]:
        return {v: k for k, v in self._obj_api_map.items()}


# ToDo: Use new syntax when requires-python >= 3.12
NO_co = TypeVar('NO_co', bound=obj_core.NotionObject, default=obj_core.NotionObject, covariant=True)


class NotionObject(Wrapper[NO_co], ABC, wraps=obj_core.NotionObject):
    """A top-level Notion API resource."""

    @property
    def id(self) -> UUID | str:
        """Return the ID of the block."""
        return raise_unset(self.obj_ref.id)

    @property
    def in_notion(self) -> bool:
        """Return whether the block was created in Notion."""
        return self.obj_ref.id != obj_core.Unset


# This acts as a simple but type-safe sentinel value, which allows narrowing by the type checker with 'is' checks.
class _Workspace(Enum):
    ROOT = 'workspace_root'


Workspace: Final = _Workspace.ROOT
"""This represents the actual root workspace in Notion."""
WorkspaceType: TypeAlias = Literal[_Workspace.ROOT]
"""This represents the type of the root workspace in Notion for type hinting."""

# ToDo: Use new syntax when requires-python >= 3.12
NE_co = TypeVar('NE_co', bound=obj_core.NotionEntity, default=obj_core.NotionEntity, covariant=True)


class NotionEntity(NotionObject[NE_co], ABC, wraps=obj_core.NotionEntity):
    def __eq__(self, other: object) -> bool:
        if other is None:
            return False
        elif not isinstance(other, NotionEntity):
            msg = f'Cannot compare {self.__class__.__name__} with {type(other).__name__}'
            raise RuntimeError(msg)
        else:
            return self.id == other.id

    def __hash__(self) -> int:
        return hash(self.id)

    @property
    def id(self) -> UUID:
        """Return the ID of the entity."""
        return raise_unset(self.obj_ref.id)

    @property
    def created_time(self) -> dt.datetime:
        """Return the time when the block was created."""
        return raise_unset(self.obj_ref.created_time)

    @property
    def created_by(self) -> User:
        """Return the user who created the block."""
        session = get_active_session()
        created_by = raise_unset(self.obj_ref.created_by)
        return session.get_user(raise_unset(created_by.id), raise_on_unknown=False)

    @property
    def last_edited_time(self) -> dt.datetime:
        """Return the time when the block was last edited."""
        return raise_unset(self.obj_ref.last_edited_time)

    @property
    def parent(self) -> NotionEntity | WorkspaceType | None:
        """Return the parent Notion entity, Workspace if the workspace is the parent, or None if not accessible."""
        session = get_active_session()
        parent = raise_unset(self.obj_ref.parent)

        match parent:
            case objs.WorkspaceRef():
                return Workspace
            case objs.PageRef(page_id=page_id):
                try:
                    return session.get_page(page_ref=page_id)
                except APIResponseError as e:
                    msg = f'Error retrieving page with id `{page_id}`: {e}'
                    _logger.warning(msg)
                    return None
            case objs.DatabaseRef(database_id=database_id):
                try:
                    return session.get_db(db_ref=database_id)
                except APIResponseError as e:
                    msg = f'Error retrieving database with id `{database_id}`: {e}'
                    _logger.warning(msg)
                    return None
            case objs.BlockRef(block_id=block_id):
                return session.get_block(block_ref=block_id)
            case _:
                msg = f'Unknown parent reference {type(parent)}'
                raise RuntimeError(msg)

    @property
    def ancestors(self) -> tuple[NotionEntity, ...]:
        """Return all ancestors from the workspace to the actual record (excluding)."""
        if (parent := self.parent) is None or parent is Workspace:
            return ()
        else:
            return (*parent.ancestors, parent)

    @property
    def is_page(self) -> bool:
        """Return whether the object is a page."""
        return False

    @property
    def is_db(self) -> bool:
        """Return whether the object is a database."""
        return False


def get_active_session() -> Session:
    """Return the current active session or raise an exception.

    Avoids cyclic imports when used within the package itself.
    For internal use mostly.
    """
    from ultimate_notion.session import Session  # noqa: PLC0415

    return Session.get_active()


def get_url(object_id: UUID | str) -> str:
    """Return the URL for the object with the given id."""
    object_id = object_id if isinstance(object_id, UUID) else UUID(object_id)
    return f'https://notion.so/{object_id.hex}'


def get_repr(obj: Any, /, *, name: Any = None, desc: Any = None) -> str:
    """Default representation, i.e. `repr(...)`, used by us for consistency."""
    type_str = str(name) if name is not None else obj.__class__.__name__
    desc_str = str(desc) if desc is not None else str(obj)
    return f"<{type_str}: '{desc_str}' at {hex(id(obj))}>"



"""Blocks that make up the content of a page.

Following blocks can be nested, i.e. they can contain children:

- Paragraph
- Headings if toggleable (*)
- Quote
- Callout
- BulletedItem
- NumberedItem
- ToDoItem
- ToggleItem
- Column within a Columns object (*)
- Table as list of TableRow objects (*)
- Synced Block
- Template (read-only)

Blocks with (*) don't have an official `children` field in the Notion API call but can still contain children using the
`append` method. They can also be created offline and populated with children offline before being pushed to Notion.
This is a technical detail and should not affect the user experience.
"""

from __future__ import annotations

import itertools
import mimetypes
from abc import ABC, abstractmethod
from collections import deque
from collections.abc import Iterator, Sequence
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, TypeGuard, cast, overload

import numpy as np
from notion_client import APIResponseError
from tabulate import tabulate
from typing_extensions import Self, TypeVar
from url_normalize import url_normalize

from ultimate_notion.comment import Comment, Discussion
from ultimate_notion.core import NotionEntity, get_active_session, get_url
from ultimate_notion.emoji import CustomEmoji, Emoji
from ultimate_notion.errors import InvalidAPIUsageError
from ultimate_notion.file import AnyFile, ExternalFile, NotionFile
from ultimate_notion.markdown import md_comment
from ultimate_notion.obj_api import blocks as obj_blocks
from ultimate_notion.obj_api import objects as objs
from ultimate_notion.obj_api.core import UnsetType, raise_unset
from ultimate_notion.obj_api.enums import BGColor, CodeLang, Color
from ultimate_notion.rich_text import Text, User
from ultimate_notion.utils import set_attr_none

if TYPE_CHECKING:
    from ultimate_notion.database import Database
    from ultimate_notion.page import Page


MIN_COLS = 2
"""Minimum number of columns when creating a column block to structure a page."""
MAX_BLOCK_CHILDREN = 100
""""The maximum number of block children that can be appended in one API call.

Source: https://developers.notion.com/reference/patch-block-children
"""
MAX_BLOCKS_PER_REQUEST = 1000
"""The maximum number of blocks that can be created/updated/deleted in one API request.

Source: https://developers.notion.com/reference/request-limits#limits-for-property-values
"""
MAX_NESTING_LEVEL = 2
"""The maximum nesting level of blocks in Notion for one append call."""

# ToDo: Use new syntax when requires-python >= 3.12
DO_co = TypeVar('DO_co', bound=obj_blocks.DataObject, default=obj_blocks.DataObject, covariant=True)


def wrap_icon(
    icon_obj: objs.FileObject | objs.EmojiObject | objs.CustomEmojiObject,
) -> NotionFile | ExternalFile | CustomEmoji | Emoji:
    """Wrap the icon object into the corresponding class."""
    match icon_obj:
        case objs.ExternalFile():
            return ExternalFile.wrap_obj_ref(icon_obj)
        case objs.HostedFile():
            return NotionFile.wrap_obj_ref(icon_obj)
        case objs.EmojiObject():
            return Emoji.wrap_obj_ref(icon_obj)
        case objs.CustomEmojiObject():
            return CustomEmoji.wrap_obj_ref(icon_obj)
        case _:
            msg = f'unknown icon object of {type(icon_obj)}'
            raise RuntimeError(msg)


class DataObject(NotionEntity[DO_co], wraps=obj_blocks.DataObject):
    """The base type for all data-related types, i.e, pages, databases and blocks."""

    @property
    def block_url(self) -> str:
        """Return the URL of the block.

        !!! note

            Databases and pages are also considered blocks in Notion but they also have a
            long-form URL. Use the `url` property to get the long-form URL.
        """
        return get_url(self.id)

    @property
    def last_edited_by(self) -> User:
        """Return the user who last edited the block."""
        session = get_active_session()
        last_edit_user_ref = raise_unset(self.obj_ref.last_edited_by)
        return session.get_user(raise_unset(last_edit_user_ref.id))

    @property
    def has_children(self) -> bool:
        """Return whether the object has children."""
        return self.obj_ref.has_children

    def _delete_me_from_parent(self) -> None:
        """Remove the block from the parent's children list."""
        if isinstance(self.parent, ChildrenMixin) and self.parent._children is not None:
            for idx, child in enumerate(self.parent._children):
                if child.id == self.id:
                    del self.parent._children[idx]
                    break
            self.parent.obj_ref.has_children = bool(self.parent._children)

    def delete(self) -> Self:
        """Delete the block.

        Pages and databases are moved to the trash, blocks are deleted permanently.
        """
        session = get_active_session()
        self.obj_ref = cast(DO_co, session.api.blocks.delete(self.id))
        self._delete_me_from_parent()
        return self

    @property
    def is_deleted(self) -> bool:
        """Return wether the object is in trash."""
        return self.obj_ref.in_trash or self.obj_ref.archived

    @abstractmethod
    def to_markdown(self) -> str:
        """Return the content of the block as Markdown."""

    def _to_markdown(self) -> str:
        """Return the content of the block as Markdown.

        This method can be overridden to provide a custom Markdown representation,
        e.g. for a page/database child block.
        """
        return self.to_markdown()


class ChildrenMixin(DataObject[DO_co], wraps=obj_blocks.DataObject):
    """Mixin for data objects that can have children

    Note that we don't use the `children` property of some Notion objects, e.g. paragraph, quote, etc.,
    as not every object has this property, e.g. a page or toggleable heading. The `children` property
    is mostly used to send a nested batch of blocks with children to Notion in one API call. It cannot
    be trusted though that the `children` property is always up-to-date, e.g. when blocks are appended
    as the Notion API often doesn't send the `children` property back in the response.
    """

    _children: list[Block] | None = None

    @property
    def _has_children_field(self) -> bool:
        """Return whether the object has a `children` field."""
        return isinstance(self.obj_ref, obj_blocks.Block) and hasattr(self.obj_ref.value, 'children')

    def _gen_children_cache(self) -> list[Block]:
        """Generate the children cache."""
        if self.is_deleted:
            msg = 'Cannot retrieve children of a deleted block from Notion.'
            raise RuntimeError(msg)

        session = get_active_session()
        child_blocks_objs = list(session.api.blocks.children.list(parent=objs.get_uuid(self.obj_ref)))
        self.obj_ref.has_children = bool(child_blocks_objs)  # update the property manually to avoid API call
        if self._has_children_field:
            self.obj_ref.value.children = child_blocks_objs  # type: ignore[attr-defined] # update the children attribute

        child_blocks: list[Block] = [Block.wrap_obj_ref(block) for block in child_blocks_objs]

        for idx, child_block in enumerate(child_blocks):
            if isinstance(child_block, ChildPage):
                child_blocks[idx] = cast(Block, child_block.page)
            elif isinstance(child_block, ChildDatabase):
                try:
                    ref_block = cast(Block, child_block.db)
                except APIResponseError:
                    # linked database that cannot be retrieved via API. Check the docs:
                    # https://developers.notion.com/reference/retrieve-a-database
                    ref_block = cast(Block, child_block)
                child_blocks[idx] = ref_block

        return [cast(Block, session.cache.setdefault(block.id, block)) for block in child_blocks]

    @property
    def children(self) -> tuple[Block, ...]:
        """Return the children of this block."""
        if self.in_notion:
            if self._children is None:
                self._children = self._gen_children_cache()
            children = tuple(self._children)  # copy list to allow deleting blocks while iterating over it
        elif self._children is not None:
            children = tuple(self._children)
        else:
            children = ()
        return children

    def append(self, blocks: Block | Sequence[Block], *, after: Block | None = None, sync: bool | None = None) -> Self:
        """Append a block or a sequence of blocks to the content of this block.

        If *this* block is already in Notion, the blocks are appended directly to Notion, otherwise
        they are prepared to be appended in one batch call, later.
        Note that explicitely defining `sync` will not affect the append behaviour itself but will raise
        an error if it is not what was expected, e.g. `sync=True` when this block is not yet in Notion.

        Args:
            blocks: A block or a sequence of blocks to append.
            after: A block to append the new blocks after.
            sync: Whether to sync the changes with Notion directly or in one batch call later.
                  If `sync = None` (default), the blocks are appended directly if this block is already in Notion,
                  otherwise they are prepared to be appended in one batch call, later.
        """
        blocks = [blocks] if isinstance(blocks, Block) else blocks

        for block in blocks:
            if not isinstance(block, Block):
                msg = f'Cannot append {type(block)} to a block.'
                raise ValueError(msg)
            if block.in_notion:
                msg = 'Cannot append a block that is already in Notion.'
                raise InvalidAPIUsageError(msg)

        if self._children is None:
            self._children = self._gen_children_cache() if self.in_notion else []
        if not blocks:
            return self

        if sync is True and not self.in_notion:
            msg = 'Cannot append blocks synchronously to a block that is not in Notion.'
            raise InvalidAPIUsageError(msg)
        elif sync is False:
            if self.in_notion:
                msg = 'Cannot append asynchronously blocks to a block that is already in Notion.'
                raise InvalidAPIUsageError(msg)
            elif after is not None:
                msg = 'Appending asynchronously blocks after a specific block is not supported.'
                raise InvalidAPIUsageError(msg)
        elif not self.in_notion:
            if after is not None:
                msg = 'Cannot append blocks after a specific block if this block is not in Notion.'
                raise InvalidAPIUsageError(msg)
            self.obj_ref.has_children = True
            for block in blocks:
                block._parent = cast(Block, self)  # set fallback parent for offline use
            # Don't store it in the potential `children` field of the obj_ref and let the Notion API do it.
            self._children.extend(blocks)
        else:  # self.in_notion and sync is not False
            blocks_iter = _chunk_blocks_for_api(cast(ParentBlock, self), blocks)
            _append_block_chunks(blocks_iter, after=after)

        return self


class CommentMixin(DataObject[DO_co], wraps=obj_blocks.DataObject):
    """Mixin for objects that can have comments and discussions."""

    _comments: list[Discussion] | None = None

    @staticmethod
    def _group_by_discussions(comments: list[Comment]) -> list[list[Comment]]:
        comments.sort(key=lambda comment: comment.discussion_id)
        grouped = [list(group) for _, group in itertools.groupby(comments, key=lambda comment: comment.discussion_id)]
        grouped.sort(key=lambda group: group[0].created_time)
        return grouped

    def _generate_comments_cache(self) -> list[Discussion]:
        """Generate the comments cache."""
        if self.is_deleted:
            msg = 'Cannot retrieve comments of a deleted block from Notion.'
            raise RuntimeError(msg)

        session = get_active_session()
        comment_objs = session.api.comments.list(self.id)
        comments = [Comment.wrap_obj_ref(comment_obj) for comment_obj in comment_objs]
        return [Discussion(comment_thread, parent=self) for comment_thread in self._group_by_discussions(comments)]

    @property
    def _discussions(self) -> tuple[Discussion, ...]:
        """Return comments of this block or page as list of discussions, i.e. threads of comments."""
        if self.in_notion:
            if self._comments is None:
                self._comments = self._generate_comments_cache()
            comments = tuple(self._comments)
        else:
            comments = ()
        return comments


# ToDo: Use new syntax when requires-python >= 3.12
B_co = TypeVar('B_co', bound=obj_blocks.Block, default=obj_blocks.Block, covariant=True)


class Block(CommentMixin[B_co], ABC, wraps=obj_blocks.Block):
    """Abstract Notion block.

    Parent class of all block types.
    """

    _parent: Block | Page | None = None  # fallback if block is not in Notion, e.g. offline assembly of blocks

    @property
    def parent(self) -> Block | Page | None:
        """Return the parent block or page, or None if not accessible."""
        from ultimate_notion.page import Page  # noqa: PLC0415  # Avoid circular import.

        if self.in_notion:
            return cast(Block | Page | None, super().parent)
        else:
            return self._parent

    @property
    def discussions(self) -> tuple[Discussion, ...]:
        """Return comments of this block as list of discussions, i.e. threads of comments."""
        return self._discussions

    def reload(self) -> Self:
        """Reload the block from the API."""
        session = get_active_session()
        self.obj_ref = cast(B_co, session.api.blocks.retrieve(self.id))
        return self

    def _update_in_notion(self, *, exclude_attrs: Sequence[str] | None = None) -> None:
        """Update the locally modified block on Notion."""
        if self.in_notion:
            session = get_active_session()
            # missing_ok=True below to cover `Heading` which behave like `TextBlock` but have no `children` attribute
            set_attr_none(self.obj_ref, exclude_attrs, inplace=True, missing_ok=True)
            session.api.blocks.update(self.obj_ref)

    def replace(self, blocks: Block | Sequence[Block]) -> None:
        """Replace this block with another block or blocks."""
        if not isinstance(blocks, Sequence):
            blocks = [blocks]

        if self.is_deleted:
            msg = 'Cannot replace a deleted block.'
            raise InvalidAPIUsageError(msg)

        for block in blocks:  # do complete sanity check first
            if block.in_notion:
                msg = f'Cannot replace with a block {block} that is already in Notion.'
                raise InvalidAPIUsageError(msg)

        if self.parent is not None and isinstance(self.parent, ChildrenMixin):
            for block in reversed(blocks):
                self.parent.append(block, after=self)
        else:
            msg = 'Cannot replace a block that has no parent.'
            raise InvalidAPIUsageError(msg)

        self.delete()

    def insert_after(self, blocks: Block | Sequence[Block]) -> None:
        """Insert a block or several blocks after this block."""
        if not isinstance(blocks, Sequence):
            blocks = [blocks]

        if self.is_deleted:
            msg = 'Cannot insert a block after a deleted block.'
            raise InvalidAPIUsageError(msg)

        for block in blocks:  # do complete sanity check first
            if block.in_notion:
                msg = f'Cannot insert block {block} that is already in Notion.'
                raise InvalidAPIUsageError(msg)

        if self.parent is not None and isinstance(self.parent, ChildrenMixin):
            self.parent.append(blocks, after=self)
        else:
            msg = 'Cannot insert a block that has no parent.'
            raise InvalidAPIUsageError(msg)

    def __eq__(self, other: Any | Block) -> bool:
        """Compare blocks by their content, not by their metadata like id, timestamp etc."""
        if not isinstance(other, Block):
            return NotImplemented
        return self.obj_ref.value == other.obj_ref.value

    def __hash__(self) -> int:
        return hash(self.obj_ref.value)


class ParentBlock(Block[B_co], ChildrenMixin[B_co], wraps=obj_blocks.Block):
    """A block that holds children blocks, mainly used for type checking.

    If there was no block like that, mypy would narrow a `Block | Page` object down to a Page object if we check
    for `isinstance(i, ChildrenMixin)` as Page inherits from ChildrenMixin and Block not. This is not what we want.
    """

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, ParentBlock):
            if isinstance(other, Block):
                return False
            return NotImplemented

        return super().__eq__(other) and self.children == other.children

    def __hash__(self) -> int:
        return hash((super().__hash__(), *self.children))

    @classmethod
    def wrap_obj_ref(cls: type[Self], obj_ref: B_co, /) -> Self:  # type: ignore[misc] # breaking covariance
        self = super().wrap_obj_ref(obj_ref)
        if obj_ref.has_children and hasattr(self.obj_ref.value, 'children') and self.obj_ref.value.children:
            self._children = [Block.wrap_obj_ref(child) for child in self.obj_ref.value.children]
            self.obj_ref.value.children = []  # clear children to avoid confusion during comparison
        return self


class CaptionMixin(Block[B_co], wraps=obj_blocks.Block):
    """Mixin for objects that can have captions."""

    @property
    def caption(self) -> Text | None:
        """Return the caption of the code block."""
        if not isinstance(self.obj_ref.value, obj_blocks.CaptionMixin | objs.FileObject):
            msg = f'Block `{type(self).__name__}` with type data `{type(self.obj_ref.value).__name__}` has no caption.'
            raise RuntimeError(msg)
        if not (caption := self.obj_ref.value.caption):
            return None
        return Text.wrap_obj_ref(caption)

    @caption.setter
    def caption(self, caption: str | None) -> None:
        if not isinstance(self.obj_ref.value, obj_blocks.CaptionMixin | objs.FileObject):
            msg = f'Block `{type(self).__name__}` with type data `{type(self.obj_ref.value).__name__}` has no caption.'
            raise RuntimeError(msg)
        self.obj_ref.value.caption = Text(caption).obj_ref if caption is not None else []
        self._update_in_notion()


# ToDo: Use new syntax when requires-python >= 3.12
TB_co = TypeVar('TB_co', bound=obj_blocks.TextBlock, default=obj_blocks.TextBlock, covariant=True)


class TextBlock(Block[TB_co], ABC, wraps=obj_blocks.TextBlock):
    """Abstract Text block.

    Parent class of all text block types.
    """

    def __init__(self, text: str) -> None:
        super().__init__()
        self.obj_ref.value.rich_text = Text(text).obj_ref

    @property
    def rich_text(self) -> Text | None:
        """Return the text content of this text block."""
        if (rich_texts := self.obj_ref.value.rich_text) is None:
            return None
        return Text.wrap_obj_ref(rich_texts)

    @rich_text.setter
    def rich_text(self, text: str | None) -> None:
        # Right now this leads to a mypy error, see
        # https://github.com/python/mypy/issues/3004
        # Always pass `RichText` objects to this setter to avoid this error.
        if text is None:
            text = ''
        self.obj_ref.value.rich_text = Text(text).obj_ref
        self._update_in_notion(exclude_attrs=['paragraph.children'])

    def __str__(self) -> str:
        """Return the text content of this block as string."""
        return self.rich_text.to_plain_text() if self.rich_text else ''

    def to_markdown(self) -> str:
        """Return the text content of this block as Markdown."""
        if self.rich_text is None:
            return ''
        return self.rich_text.to_markdown()


class Code(TextBlock[obj_blocks.Code], CaptionMixin[obj_blocks.Code], wraps=obj_blocks.Code):
    """Code block."""

    def __init__(
        self,
        text: str,
        *,
        language: str | CodeLang = CodeLang.PLAIN_TEXT,
        caption: str | None = None,
    ) -> None:
        super().__init__(text)
        if not isinstance(language, CodeLang):
            language = CodeLang(language)
        self.obj_ref.value.language = language
        self.obj_ref.value.caption = Text(caption).obj_ref if caption is not None else []

    @property
    def language(self) -> CodeLang:
        """Return the programming language of the code block."""
        return self.obj_ref.value.language

    @language.setter
    def language(self, language: CodeLang) -> None:
        self.obj_ref.value.language = language
        self._update_in_notion()

    def to_markdown(self) -> str:
        """Return the code content of this block as Markdown."""
        lang = self.obj_ref.value.language
        return f'```{lang}\n{super().to_markdown()}\n```'


# ToDo: Use new syntax when requires-python >= 3.12
TB = TypeVar('TB', bound=obj_blocks.ColoredTextBlock)


class ColoredTextBlock(TextBlock[TB], wraps=obj_blocks.ColoredTextBlock):
    """Abstract Text block with color.

    Parent class of all text block types with color.
    """

    def __init__(
        self,
        text: str,
        *,
        color: Color | BGColor = Color.DEFAULT,
    ) -> None:
        super().__init__(text)
        value = self._get_value()
        value.color = color

    def _get_value(self) -> obj_blocks.ColoredTextBlockTypeData:
        return self.obj_ref.value

    @property
    def color(self) -> Color | BGColor:
        """Return the color of the text block."""
        return self._get_value().color

    @color.setter
    def color(self, color: Color | BGColor) -> None:
        self._get_value().color = color
        self._update_in_notion(exclude_attrs=['paragraph.children'])


class Paragraph(ColoredTextBlock[obj_blocks.Paragraph], ParentBlock[obj_blocks.Paragraph], wraps=obj_blocks.Paragraph):
    """Paragraph block."""


# ToDo: Use new syntax when requires-python >= 3.12
HT = TypeVar('HT', bound=obj_blocks.Heading)


class Heading(ColoredTextBlock[HT], ParentBlock[obj_blocks.Heading], wraps=obj_blocks.Heading):
    """Abstract Heading block.

    Parent class of all heading block types.
    """

    def __init__(
        self,
        text: str,
        *,
        color: Color | BGColor = Color.DEFAULT,
        toggleable: bool = False,
    ) -> None:
        super().__init__(text, color=color)
        self.obj_ref.value.is_toggleable = toggleable

    @property
    def toggleable(self) -> bool:
        """Return whether the heading is toggleable."""
        return self.obj_ref.value.is_toggleable

    @toggleable.setter
    def toggleable(self, toggleable: bool) -> None:
        """Set the heading to be toggleable or not."""
        if self.toggleable and self.has_children:
            msg = 'Cannot make a toggleable heading non-toggleable if it has children.'
            raise InvalidAPIUsageError(msg)
        self.obj_ref.value.is_toggleable = toggleable
        self._update_in_notion()

    def append(self, blocks: Block | Sequence[Block], *, after: Block | None = None, sync: bool | None = None) -> Self:
        if not self.toggleable:
            msg = 'Cannot append blocks to a non-toggleable heading.'
            raise InvalidAPIUsageError(msg)
        return super().append(blocks, after=after, sync=sync)


class Heading1(Heading[obj_blocks.Heading1], wraps=obj_blocks.Heading1):
    """Heading 1 block."""

    def to_markdown(self) -> str:
        return f'## {super().to_markdown()}'  # we use ## as # is used for page titles


class Heading2(Heading[obj_blocks.Heading2], wraps=obj_blocks.Heading2):
    """Heading 2 block."""

    def to_markdown(self) -> str:
        return f'### {super().to_markdown()}'


class Heading3(Heading[obj_blocks.Heading3], wraps=obj_blocks.Heading3):
    """Heading 3 block."""

    def to_markdown(self) -> str:
        return f'#### {super().to_markdown()}'


class Quote(ColoredTextBlock[obj_blocks.Quote], ParentBlock[obj_blocks.Quote], wraps=obj_blocks.Quote):
    """Quote block."""

    def to_markdown(self) -> str:
        return f'> {super().to_markdown()}\n'


class Callout(ColoredTextBlock[obj_blocks.Callout], ParentBlock[obj_blocks.Callout], wraps=obj_blocks.Callout):
    """Callout block.

    !!! note

        The default icon is an electric light bulb, i.e. 💡 in case `None` is passed as icon.
        It is not possible to remove the icon, but you can change it to a different emoji or a file.
    """

    def __init__(
        self,
        text: str,
        *,
        color: Color | BGColor = Color.DEFAULT,
        icon: AnyFile | str | Emoji | CustomEmoji | None = None,
    ) -> None:
        super().__init__(text, color=color)
        if isinstance(icon, str) and not isinstance(icon, Emoji | CustomEmoji):
            icon = Emoji(icon)
        if icon is None:
            # Use the Notion default icon instead of sending Unset, which would be replaced by Notion with their
            # default anyway. This will get us into trouble if Notion ever changes their default icon. But for now,
            # this is the best we can do for comparison etc.
            icon = self.get_default_icon()

        self.obj_ref.value.icon = icon.obj_ref

    @staticmethod
    def get_default_icon() -> Emoji:
        """Return the default icon of a callout block."""
        return Emoji('💡')

    @property
    def icon(self) -> NotionFile | ExternalFile | Emoji | CustomEmoji:
        if isinstance(icon := self.obj_ref.value.icon, UnsetType):
            return self.get_default_icon()
        else:
            return wrap_icon(icon)

    @icon.setter
    def icon(self, icon: AnyFile | Emoji | CustomEmoji | None) -> None:
        if icon is None:
            icon = self.get_default_icon()
        self.obj_ref.value.icon = icon.obj_ref
        self._update_in_notion()

    def to_markdown(self) -> str:
        match self.icon:
            case Emoji():
                return f'{self.icon} {super().to_markdown()}\n'
            case CustomEmoji():
                return f':{self.icon.name}: {super().to_markdown()}\n'
            case AnyFile():
                return f'![icon]({self.icon.url}) {super().to_markdown()}\n'
            case _:
                msg = f'Invalid icon type {type(self.icon)}'
                raise ValueError(msg)


class BulletedItem(
    ColoredTextBlock[obj_blocks.BulletedListItem],
    ParentBlock[obj_blocks.BulletedListItem],
    wraps=obj_blocks.BulletedListItem,
):
    """Bulleted list item."""

    def to_markdown(self) -> str:
        return f'- {super().to_markdown()}\n'


class NumberedItem(
    ColoredTextBlock[obj_blocks.NumberedListItem],
    ParentBlock[obj_blocks.NumberedListItem],
    wraps=obj_blocks.NumberedListItem,
):
    """Numbered list item."""

    def to_markdown(self) -> str:
        return f'1. {super().to_markdown()}\n'


class ToDoItem(ColoredTextBlock[obj_blocks.ToDo], ParentBlock[obj_blocks.ToDo], wraps=obj_blocks.ToDo):
    """ToDo list item."""

    def __init__(
        self,
        text: str,
        *,
        checked: bool = False,
        color: Color | BGColor = Color.DEFAULT,
    ) -> None:
        super().__init__(text, color=color)
        self.obj_ref.value.checked = checked

    @property
    def checked(self) -> bool:
        return self.obj_ref.to_do.checked

    @checked.setter
    def checked(self, checked: bool) -> None:
        self.obj_ref.value.checked = checked
        self._update_in_notion()

    def to_markdown(self) -> str:
        mark = 'x' if self.checked else ' '
        return f'- [{mark}] {super().to_markdown()}\n'


class ToggleItem(ColoredTextBlock[obj_blocks.Toggle], ParentBlock[obj_blocks.Toggle], wraps=obj_blocks.Toggle):
    """Toggle list item."""

    def to_markdown(self) -> str:
        return f'- {super().to_markdown()}\n'


class Divider(Block[obj_blocks.Divider], wraps=obj_blocks.Divider):
    """Divider block."""

    def to_markdown(self) -> str:  # noqa: PLR6301
        return '---\n'


class TableOfContents(Block[obj_blocks.TableOfContents], wraps=obj_blocks.TableOfContents):
    """Table of Contents block."""

    def __init__(self, *, color: Color | BGColor = Color.DEFAULT) -> None:
        super().__init__()
        self.obj_ref.value.color = color

    def to_markdown(self) -> str:  # noqa: PLR6301
        return '```{toc}\n```'


class Breadcrumb(Block[obj_blocks.Breadcrumb], wraps=obj_blocks.Breadcrumb):
    """Breadcrumb block."""

    def to_markdown(self) -> str:
        def is_page(obj: NotionEntity) -> TypeGuard[Page]:
            return obj.is_page

        return ' / '.join(ancestor.title or 'Untitled Page' for ancestor in self.ancestors if is_page(ancestor)) + '\n'


class Embed(CaptionMixin[obj_blocks.Embed], wraps=obj_blocks.Embed):
    """Embed block."""

    def __init__(self, url: str, *, caption: str | None = None) -> None:
        caption_obj = Text(caption).obj_ref if caption is not None else None
        super().__init__(url=url_normalize(url), caption=caption_obj)

    @property
    def url(self) -> str:
        """Return the URL of the embedded item."""
        return self.obj_ref.value.url

    @url.setter
    def url(self, url: str) -> None:
        if (norm_url := url_normalize(url)) is not None:
            self.obj_ref.value.url = norm_url
            self._update_in_notion()
        else:
            msg = f'Invalid URL: {url}'
            raise ValueError(msg)

    def to_markdown(self) -> str:
        title = self.caption.to_plain_text() if self.caption is not None else self.url
        if self.url is not None:
            return f'[{title}]({self.url})\n'
        else:
            return ''


class Bookmark(CaptionMixin[obj_blocks.Bookmark], wraps=obj_blocks.Bookmark):
    """Bookmark block."""

    def __init__(self, url: str, *, caption: str | None = None) -> None:
        caption_obj = Text(caption).obj_ref if caption is not None else None
        super().__init__(url=url_normalize(url), caption=caption_obj)

    @property
    def url(self) -> str | None:
        """Return the URL of the bookmark."""
        if url := self.obj_ref.value.url:
            return url
        return None

    @url.setter
    def url(self, url: str | None) -> None:
        url = url_normalize(url)
        self.obj_ref.value.url = '' if url is None else url
        self._update_in_notion()

    def to_markdown(self) -> str:
        title = self.caption.to_plain_text() if self.caption is not None else self.url
        if self.url is not None:
            return f'Bookmark: [{title}]({self.url})\n'
        else:
            return 'Bookmark: [Add a web bookmark]()\n'  # emtpy bookmark


class LinkPreview(Block[obj_blocks.LinkPreview], wraps=obj_blocks.LinkPreview):
    """Link preview block.

    !!! warning "Not Supported"

        The `link_preview` block can only be returned as part of a response.
        The Notion API does not support creating or appending `link_preview` blocks.
    """

    def __init__(self, url: str) -> None:
        msg = 'The Notion API does not support creating or appending `link_preview` blocks.'
        raise NotImplementedError(msg)
        # ToDo: Implement this when the API supports it
        # super().__init__()
        # self.obj_ref.value.url = url

    @property
    def url(self) -> str | None:
        return self.obj_ref.value.url

    def to_markdown(self) -> str:
        if self.block_url is not None:
            return f'Link preview: [{self.url}]({self.url})\n'
        return super().to_markdown()


class Equation(Block[obj_blocks.Equation], wraps=obj_blocks.Equation):
    """Equation block.

    LaTeX equation in display mode, e.g. `$$ \\mathrm{E=mc^2} $$`, but without the `$$` signs.
    """

    def __init__(self, latex: str) -> None:
        super().__init__()
        self.obj_ref.value.expression = latex

    @property
    def latex(self) -> str:
        """Return the LaTeX expression of the equation."""
        if (expr := self.obj_ref.equation.expression) is not None:
            return expr.rstrip()
        else:
            return ''

    @latex.setter
    def latex(self, latex: str) -> None:
        self.obj_ref.value.expression = latex
        self._update_in_notion()

    def to_markdown(self) -> str:
        """Return the LaTeX expression of the equation as Markdown."""
        return f'$$\n{self.latex}\n$$\n'


FT = TypeVar('FT', bound=obj_blocks.FileBase)


class FileBaseBlock(CaptionMixin[FT], ABC, wraps=obj_blocks.FileBase):
    """Abstract Block for file-based blocks.

    Parent class of all file-based block types.
    """

    def __init__(
        self,
        file: AnyFile,
        *,
        caption: str | None = None,
        name: str | None = None,
    ) -> None:
        super().__init__()
        if isinstance(file, str):
            msg = 'Expected a file object! Use `uno.url` or upload a local file first.'
            raise ValueError(msg)

        file.name = name if name is not None else file.name
        file.caption = caption if caption is not None else file.caption
        self.obj_ref.value = file.obj_ref

    @property
    def file_info(self) -> AnyFile:
        """Return the file information of this block as a copy."""
        if isinstance(file_obj := self.obj_ref.value, objs.FileObject):
            return AnyFile.wrap_obj_ref(file_obj.model_copy(deep=True))
        else:
            msg = f'Unknown file type {type(file_obj)}'
            raise ValueError(msg)

    @property
    def url(self) -> str:
        """Return the URL of the file."""
        if isinstance(file_info := self.file_info, NotionFile | ExternalFile):
            return file_info.url
        else:
            msg = f'File type {type(file_info)} has no URL.'
            raise ValueError(msg)


class File(FileBaseBlock[obj_blocks.File], wraps=obj_blocks.File):
    """File block.

    !!! note

        Only the caption and name can be modified, the file object is read-only.
        Note that only the file name not the file suffix can be modified,
        the suffix is determined initially by the url.
    """

    def __init__(
        self,
        file: AnyFile,
        *,
        name: str | None = None,
        caption: str | None = None,
    ) -> None:
        super().__init__(file, caption=caption, name=name)

    @property
    def name(self) -> str:
        """Return the name of the file."""
        name = self.file_info.name
        return name if name is not None else ''

    @name.setter
    def name(self, name: str) -> None:
        self.obj_ref.value.name = name
        self._update_in_notion()

    def to_markdown(self) -> str:
        """Return the file link as Markdown."""
        md = f'[📎 {self.name}]({self.url})\n'
        if self.caption is not None:
            md += f'{self.caption.to_markdown()}\n'
        return md


class Image(FileBaseBlock[obj_blocks.Image], wraps=obj_blocks.Image):
    """Image block.

    !!! note

        Only the caption can be modified, the URL is read-only.
    """

    def __init__(self, file: AnyFile, *, caption: str | None = None) -> None:
        super().__init__(file, caption=caption)

    def to_markdown(self) -> str:
        """Return the image as Markdown."""
        alt = self.url.rsplit('/').pop()
        if self.caption is not None:
            caption = self.caption.to_plain_text()
            return f'<figure><img src="{self.url}" alt="{alt}" /><figcaption>{caption}</figcaption></figure>\n'
        else:
            return f'![{alt}]({self.url})\n'


class PDF(FileBaseBlock[obj_blocks.PDF], wraps=obj_blocks.PDF):
    """PDF block.

    !!! note

        Only the caption can be modified, the URL is read-only.
    """

    def __init__(self, file: AnyFile, *, caption: str | None = None) -> None:
        file.name = None  # PDF files cannot have a name
        super().__init__(file, caption=caption)

    def to_markdown(self) -> str:
        """Return the PDF as Markdown."""
        name = self.url.rsplit('/').pop()
        md = f'[📖 {name}]({self.url})\n'
        if self.caption:
            md += f'{self.caption.to_markdown()}\n'
        return md


class Video(FileBaseBlock[obj_blocks.Video], wraps=obj_blocks.Video):
    """Video block.

    !!! note

        Only the caption can be modified, the URL is read-only.
    """

    def __init__(self, file: AnyFile, *, caption: str | None = None) -> None:
        file.name = None  # Video files cannot have a name
        super().__init__(file, caption=caption)

    def to_markdown(self) -> str:
        """Return the video as Markdown."""
        mime_type, _ = mimetypes.guess_type(self.url)
        vtype = f' type="{mime_type}"' if mime_type else ''
        caption = '' if self.caption is None else self.caption.to_markdown()
        md = f'<video width="320" height="240" controls><source src="{self.url}"{vtype}>{caption}</video>\n'
        return md


class Audio(FileBaseBlock[obj_blocks.Audio], wraps=obj_blocks.Audio):
    """Audio block.

    !!! note

        Only the caption can be modified, the URL is read-only.
    """

    def __init__(self, file: AnyFile, *, caption: str | None = None) -> None:
        file.name = None  # Audio files cannot have a name
        super().__init__(file, caption=caption)

    def to_markdown(self) -> str:
        """Return the audio as Markdown."""
        mime_type, _ = mimetypes.guess_type(self.url)
        atype = f' type="{mime_type}"' if mime_type else ''
        caption = '' if self.caption is None else self.caption.to_markdown()
        md = f'<audio controls><source src="{self.url}"{atype}>{caption}</audio>\n'
        return md


class ChildPage(Block[obj_blocks.ChildPage], wraps=obj_blocks.ChildPage):
    """Child page block.

    !!! note

        To create a child page block, create a new page with the corresponding parent.
        This block is used only internally.
    """

    def __init__(self) -> None:
        msg = 'To create a child page block, create a new page with the corresponding parent.'
        raise InvalidAPIUsageError(msg)

    @property
    def title(self) -> str | None:
        """Return the title of the child page."""
        if title := raise_unset(self.obj_ref.child_page.title):
            return title
        return None

    @property
    def page(self) -> Page:
        """Return the actual Page object."""
        sess = get_active_session()
        return sess.get_page(self.id)

    def to_markdown(self) -> str:  # noqa: PLR6301
        msg = '`ChildPage` is only used internally, work with a proper `Page` instead.'
        raise InvalidAPIUsageError(msg)


class ChildDatabase(Block[obj_blocks.ChildDatabase], wraps=obj_blocks.ChildDatabase):
    """Child database block.

    This block is used to represent a database if it is a child of e.g. a page.
    We try to resolve it via the API to get the actual database object. This does not work
    if it is a `linked database` as mentioned in the Notion API docs:
    https://developers.notion.com/reference/retrieve-a-database

    !!! note

        To create a child database block as an end-user, create a new database with the corresponding parent.
        This block is used only internally.
    """

    def __init__(self) -> None:
        msg = 'To create a child database block, create a new database with the corresponding parent.'
        raise InvalidAPIUsageError(msg)

    @property
    def title(self) -> str | None:
        """Return the title of the child database"""
        if title := raise_unset(self.obj_ref.child_database.title):
            return title
        return None

    @property
    def db(self) -> Database:
        """Return the actual Database object."""
        sess = get_active_session()
        return sess.get_db(self.id)

    def to_markdown(self) -> str:  # noqa: PLR6301
        """Return the child database as Markdown."""
        return '<kbd>↗️ Linked database (unsupported)</kbd>'


class Column(ParentBlock[obj_blocks.Column], wraps=obj_blocks.Column):
    """Column block."""

    def __init__(self) -> None:
        msg = 'Column blocks cannot be created directly. Use `Columns` instead.'
        raise InvalidAPIUsageError(msg)

    def to_markdown(self) -> str:
        """Return the content of this column as Markdown."""
        mds = []
        for block in self.children:
            mds.append(block.to_markdown())
        return '\n'.join(mds)

    @property
    def width_ratio(self) -> float | None:
        """Return the width ratio of this column."""
        return self.obj_ref.column.width_ratio


class Columns(ParentBlock[obj_blocks.ColumnList], wraps=obj_blocks.ColumnList):
    """Columns block holding multiple `Column` blocks.

    This block is used to create a layout with multiple columns in a single page.
    Either specify the number of columns as an integer or provide a sequence of width ratios,
    which can be positive integers or floats.
    """

    def __init__(self, columns: int | Sequence[float | int]) -> None:
        """Create a new `Columns` block with the given number of columns."""
        super().__init__()
        match columns:
            case int() as n_columns:
                if n_columns < MIN_COLS:
                    msg = f'Number of columns must be at least {MIN_COLS}.'
                    raise ValueError(msg)
                self.obj_ref.column_list.children = [obj_blocks.Column.build() for _ in range(n_columns)]
            case ratios if isinstance(ratios, Sequence) and all(isinstance(x, float | int) for x in ratios):
                if not ratios or any(r <= 0 for r in ratios):
                    msg = 'Ratios must be a non-empty sequence of positive numbers.'
                    raise ValueError(msg)
                ratios_arr = np.array(ratios, dtype=float)
                ratios_arr /= ratios_arr.sum()  # normalize ratios to sum to 1 as asked by the Notion API
                self.obj_ref.column_list.children = [obj_blocks.Column.build(width_ratio=ratio) for ratio in ratios_arr]
            case _:
                msg = 'Columns must be initialized with an integer or a sequence of floats/integers.'
                raise TypeError(msg)

    def __getitem__(self, index: int) -> Column:
        return cast(Column, self.children[index])

    def add_column(self, index: int | None = None) -> Self:
        """Add a new column to this block of columns at the given index.

        The index must be between 0 and the number of columns (inclusive).
        If no index is given, the new column is added at the end.

        To specify the width ratio of the new column, use the `width_ratios` property.
        """
        if index is None:
            index = len(self.children)
        if 0 <= index <= len(self.children):
            new_col = Column.wrap_obj_ref(obj_blocks.Column.build())
            super().append(new_col, after=self.children[index - 1] if index > 0 else None)
            return self
        else:
            msg = f'Column index must be between 0 and {len(self.children)} (inclusive).'
            raise IndexError(msg)

    def append(self, blocks: Block | Sequence[Block], *, after: Block | None = None, sync: bool | None = None) -> Self:  # noqa: PLR6301
        """Append a block or a sequence of blocks to the content of this block."""
        msg = 'Use `add_column` to append a new column.'
        raise InvalidAPIUsageError(msg)

    def to_markdown(self) -> str:
        """Return the content of all columns as Markdown."""
        cols = []
        for i, block in enumerate(self.children):
            md = md_comment(f'column {i + 1}')
            cols.append(md + block.to_markdown())
        return '\n'.join(cols)

    @property
    def width_ratios(self) -> tuple[float | None, ...]:
        """Return the width ratios of the columns."""
        return tuple(cast(Column, col).width_ratio for col in self.children)

    @width_ratios.setter
    def width_ratios(self, ratios: Sequence[float | int]) -> None:
        """Set the width ratios of the columns."""
        if len(ratios) != len(self.children):
            msg = f'Ratios must have the same length as the number of columns ({len(self.children)}).'
            raise ValueError(msg)

        if not isinstance(ratios, Sequence) or not all(isinstance(x, float | int) for x in ratios):
            msg = 'Ratios must be a sequence of floats or integers.'
            raise TypeError(msg)
        if not ratios or any(r <= 0 for r in ratios):
            msg = 'Ratios must be a non-empty sequence of positive numbers.'
            raise ValueError(msg)

        ratios_arr = np.array(ratios, dtype=float)
        ratios_arr /= ratios_arr.sum()  # normalize ratios to sum to 1 as asked by the Notion API
        for col, ratio in zip(self.children, ratios_arr, strict=False):
            obj_ref = cast(obj_blocks.Column, col.obj_ref)
            obj_ref.column.width_ratio = ratio
            col._update_in_notion(exclude_attrs=['column.children'])


class TableRow(tuple[Text | None, ...], Block[obj_blocks.TableRow], wraps=obj_blocks.TableRow):
    """Table row block behaving like a tuple."""

    def __new__(cls, *cells: str | None) -> TableRow:
        return tuple.__new__(cls, cells)

    def __init__(self, *cells: str | None) -> None:
        super().__init__(n_cells=len(cells))
        for idx, cell in enumerate(cells):
            if cell is None:
                cell = ''
            self.obj_ref.table_row.cells[idx] = Text(cell).obj_ref

    @classmethod
    def wrap_obj_ref(cls, obj_ref: obj_blocks.TableRow, /) -> TableRow:
        row = TableRow(*[Text.wrap_obj_ref(cell) if cell else None for cell in obj_ref.table_row.cells])
        row.obj_ref = obj_ref
        return row

    def to_markdown(self) -> str:
        """Return the row as Markdown."""
        return ' | '.join(s or '' for s in self)  # convert None to ''


class Table(ParentBlock[obj_blocks.Table], wraps=obj_blocks.Table):
    """Table block."""

    def __init__(self, n_rows: int, n_cols: int, *, header_col: bool = False, header_row: bool = False) -> None:
        if n_rows < 1:
            msg = 'Table must have at least one row.'
            raise InvalidAPIUsageError(msg)
        if not (1 <= n_cols <= MAX_BLOCK_CHILDREN):
            msg = f'Number of columns must be between 1 and {MAX_BLOCK_CHILDREN}.'
            raise InvalidAPIUsageError(msg)

        super().__init__()
        self.obj_ref.table.table_width = n_cols
        self.obj_ref.table.has_column_header = header_row
        self.obj_ref.table.has_row_header = header_col
        self.obj_ref.has_children = True
        self._children = [TableRow.wrap_obj_ref(obj_blocks.TableRow.build(n_cols)) for _ in range(n_rows)]

    def _check_index(self, index: int | tuple[int, int]) -> tuple[int, int | None]:
        if isinstance(index, tuple):
            row_idx, col_idx = index
        else:
            row_idx, col_idx = index, None

        if not (0 <= row_idx < len(self.children)):
            msg = 'Row index out of range'
            raise IndexError(msg)

        if col_idx is not None and not (0 <= col_idx < self.width):
            msg = 'Column index out of range'
            raise IndexError(msg)

        return row_idx, col_idx

    def __str__(self) -> str:
        return self.to_markdown()

    @overload
    def __getitem__(self, index: int) -> TableRow: ...

    @overload
    def __getitem__(self, index: tuple[int, int]) -> Text | None: ...

    def __getitem__(self, index: int | tuple[int, int]) -> Text | TableRow | None:
        row_idx, col_idx = self._check_index(index)
        row = self.children[row_idx]
        return row if col_idx is None else row[col_idx]

    def __setitem__(
        self,
        index: int | tuple[int, int],
        value: str | Sequence[str | None] | None,
    ) -> None:
        row_idx, col_idx = self._check_index(index)
        row = self.children[row_idx]
        row_obj = row.obj_ref
        if col_idx is None:
            if not isinstance(value, Sequence) or len(value) != self.width:
                msg = 'Value is no sequence or its length does not match the width of the table.'
                raise ValueError(msg)
            if isinstance(value, str):
                msg = 'A sequence of values is needed to set a whole row.'
                raise ValueError(msg)
            for idx, item in enumerate(value):
                item = item if item is not None else ''
                row_obj.table_row.cells[idx] = Text(item).obj_ref
        elif isinstance(value, str):
            value = value if value is not None else ''
            row_obj.table_row.cells[col_idx] = Text(value).obj_ref
        else:
            msg = 'A single value is needed to set a single cell.'
            raise ValueError(msg)
        row._update_in_notion()
        if self._children is not None:  # update cache
            self._children[row_idx] = TableRow.wrap_obj_ref(row_obj)  # needed to call __new__ on the TableRow

    @property
    def children(self) -> tuple[TableRow, ...]:
        """Return all rows of the table."""
        return tuple(cast(TableRow, block) for block in super().children)

    @property
    def width(self) -> int:
        """Return the width, i.e. number of columns, of the table."""
        return self.obj_ref.table.table_width

    @property
    def shape(self) -> tuple[int, int]:
        """Return the shape of the table."""
        n_cols = self.width
        n_rows = len(self.children)
        return n_rows, n_cols

    @property
    def has_header_row(self) -> bool:
        """Return whether the table has a header row."""
        return self.obj_ref.table.has_column_header

    @has_header_row.setter
    def has_header_row(self, header_row: bool) -> None:
        self.obj_ref.table.has_column_header = header_row
        self._update_in_notion(exclude_attrs=['table.table_width', 'table.children'])

    @property
    def has_header_col(self) -> bool:
        """Return whether the table has a header column."""
        return self.obj_ref.table.has_row_header

    @has_header_col.setter
    def has_header_col(self, header_col: bool) -> None:
        self.obj_ref.table.has_row_header = header_col
        self._update_in_notion(exclude_attrs=['table.table_width', 'table.children'])

    def to_markdown(self) -> str:
        """Return the table as Markdown."""
        headers = 'firstrow' if self.has_header_row else [''] * self.width
        return tabulate(self.children, headers, tablefmt='github') + '\n'

    def insert_row(self, index: int, values: Sequence[str]) -> Self:
        """Insert a new row at the given index."""
        if not (1 <= index <= len(self.children)):
            msg = f'Columns can only be inserted from index 1 to {len(self.children)}.'
            raise IndexError(msg)
        if len(values) != self.width:
            msg = 'Length of passed value does not match width of table.'
            raise ValueError(msg)

        self.append(TableRow(*values), after=self.children[index - 1])
        return self

    def append_row(self, values: Sequence[str]) -> Self:
        """Append a new row to the table."""
        return self.insert_row(len(self.children), values)


class LinkToPage(Block[obj_blocks.LinkToPage], wraps=obj_blocks.LinkToPage):
    """Link to page block.

    !!! note
        Updating a link to page block is not supported by the Notion API.
        Use `.replace(new_block)` instead.
    """

    def __init__(self, page: Page) -> None:
        page_ref = objs.PageRef.build(page.obj_ref)
        super().__init__(link_to_page=page_ref)

    @property
    def page(self) -> Page:
        """Return the page this block links to."""
        session = get_active_session()
        return session.get_page(objs.get_uuid(self.obj_ref.link_to_page))

    @page.setter
    def page(self, page: Page) -> None:  # noqa: PLR6301
        # Updating a link to page block is not supported by the API and returns the current one.
        # self.obj_ref.link_to_page = objs.PageRef.build(page.obj_ref)
        # self._update_in_notion()
        msg = 'Updating a link to page block is not supported by the API. Use `.replace(new_block)` instead.'
        raise InvalidAPIUsageError(msg)

    def to_markdown(self) -> str:
        """ "Return the link to page block as Markdown."""
        return f'[**↗️ <u>{self.page.title}</u>**]({self.page.url})\n'


class SyncedBlock(ParentBlock[obj_blocks.SyncedBlock], wraps=obj_blocks.SyncedBlock):
    """Synced block - either original or synced."""

    def __init__(self, blocks: Block | Sequence[Block]) -> None:
        """Create the original synced block."""
        super().__init__()
        blocks = [blocks] if isinstance(blocks, Block) else blocks
        self.obj_ref.synced_block.children = [block.obj_ref for block in blocks]

    @property
    def is_original(self) -> bool:
        """Return if this block is the original block."""
        return self.obj_ref.synced_block.synced_from is None

    @property
    def is_synced(self) -> bool:
        """Return if this block is synced from another block."""
        return not self.is_original

    def get_original(self) -> SyncedBlock:
        """Return the original block."""
        if self.is_original:
            return self
        elif (synced_from := self.obj_ref.synced_block.synced_from) is not None:
            session = get_active_session()
            return cast(SyncedBlock, session.get_block(objs.get_uuid(synced_from)))
        else:
            msg = 'Unknown synced block, neither original nor synced!'
            raise RuntimeError(msg)

    def create_synced(self) -> SyncedBlock:
        """Return the synced block for appending."""
        if not self.in_notion:
            msg = 'Cannot create a synced block for a block that is not in Notion. Append first!'
            raise RuntimeError(msg)

        if not self.is_original:
            msg = 'Cannot create a synced block for a block that is already synced.'
            raise RuntimeError(msg)

        obj = obj_blocks.SyncedBlock.build()
        obj.synced_block.synced_from = obj_blocks.BlockRef.build(self.obj_ref)
        return self.wrap_obj_ref(obj)

    def to_markdown(self, *, with_comment: bool = True) -> str:
        """Return the content of this synced block as Markdown."""
        if self.is_original:
            md = md_comment('original block') if with_comment else ''
            md += '\n'.join([child.to_markdown() for child in self.children])
        else:
            md = md_comment('synced block') if with_comment else ''
            md += self.get_original().to_markdown(with_comment=False)
        return md


class Template(TextBlock[obj_blocks.Template], ParentBlock[obj_blocks.Template], wraps=obj_blocks.Template):
    """Template block.

    !!! warning "Deprecated"

        As of March 27, 2023 creation of template blocks will no longer be supported.
    """

    def __init__(self) -> None:
        msg = 'A template block cannot be created by a user.'
        raise InvalidAPIUsageError(msg)

    def to_markdown(self) -> str:
        """Return the template content as Markdown."""
        return f'<button type="button">{super().to_markdown()}</button>\n'


class Unsupported(Block[obj_blocks.UnsupportedBlock], wraps=obj_blocks.UnsupportedBlock):
    """Unsupported block in the API.

    Some blocks like buttons, AI blocks, or templates are not supported by the Notion API.
    They will be returned as `Unsupported` blocks when fetched, but cannot be created or modified.
    """

    def __init__(self) -> None:
        msg = 'An unsupported block cannot be created by a user.'
        raise InvalidAPIUsageError(msg)

    def to_markdown(self) -> str:  # noqa: PLR6301
        """Return a placeholder for unsupported blocks."""
        return '<kbd>Unsupported block</kbd>\n'


def traverse_blocks(blocks: Sequence[Block]) -> Iterator[Block]:
    """Recursively traverse blocks and their children in depth-first order."""
    for block in blocks:
        yield block
        if block.has_children and isinstance(block, ChildrenMixin):
            yield from traverse_blocks(block.children)


@dataclass
class _Node:
    """Tree of nodes used to chunk blocks for the Notion API."""

    block: Block | Page
    is_root: bool = False
    children: list[_Node] = field(default_factory=list)


def _chunk_blocks_for_api(parent: Block | Page, blocks: Sequence[Block]) -> Iterator[_Node]:
    """Yield batches of blocks with a parent so that each batch fulfills the Notion API requirements.

    Requirements:
    - Maximum of 1000 blocks per request
    - Maximum of 100 blocks in an array, also when appending children of a block
    - Maximum nesting level of 2, i.e. great grandchildren are not allowed

    We won't adhere to the 500KB limit per request here, as this is hard to check without
    knowing the exact size of the serialized request. If you hit this limit, reduce the number
    of blocks you append at once. This is quite unlikely to happen in practice as the current implementation
    appends only single-level blocks at a time adhering to the 100 blocks per array limit.

    Source: https://developers.notion.com/reference/request-limits#limits-for-property-values
    """
    # Breadth-first traversal to create batches of blocks which handles special blocks like columns and tables correctly
    # Note that there are a few edge cases where this can still fail, e.g. a table with >100 rows or many tables with
    # a total of more than 1000 rows/blocks. These cases are quite unlikely in practice.
    batch = _Node(block=parent, is_root=True)
    batch_blocks = [_Node(block=block) for block in blocks]
    queue = deque([(batch, batch_blocks)])
    while queue:
        batch, batch_blocks = queue.popleft()
        if not batch_blocks:
            continue

        curr_nodes, next_nodes = batch_blocks[:MAX_BLOCK_CHILDREN], batch_blocks[MAX_BLOCK_CHILDREN:]
        batch.children.extend(curr_nodes)

        for node in curr_nodes:
            match node.block:
                case Columns() as cols:
                    col_nodes = [_Node(block=col) for col in cols.children]
                    node.children.extend(col_nodes)
                    for col_node in col_nodes:
                        queue.append(
                            (col_node, [_Node(block=child) for child in cast(Column, col_node.block).children])
                        )
                case Table() as table:
                    node.children.extend([_Node(block=row) for row in table.children])
                case ParentBlock() as parent_block if parent_block.has_children:
                    queue.append((node, [_Node(block=child) for child in parent_block.children]))
        if next_nodes:
            queue.append((_Node(block=batch.block, is_root=batch.is_root), next_nodes))
        yield batch


def _build_obj_ref(node: _Node) -> Block:
    """Recursively build the obj_ref of a block and its children."""
    block = node.block
    children = node.children
    if not isinstance(block, Block):
        msg = f'Non-block type {type(block)} not allowed on this level of the hierarchy.'
        raise TypeError(msg)
    if isinstance(block, ParentBlock) and block.has_children and hasattr(block.obj_ref.value, 'children'):
        for child in node.children:
            _build_obj_ref(child)
        block.obj_ref.value.children = [child.block.obj_ref for child in children]
    return block


def _append_block_chunks(batch_trees: Iterator[_Node], *, after: Block | None = None) -> None:
    """Append chunks of blocks to a parent block, respecting API limits."""
    session = get_active_session()
    curr_after = after

    for parent_node in batch_trees:
        parent = parent_node.block
        blocks = [_build_obj_ref(child) for child in parent_node.children]

        after = curr_after if parent_node.is_root and curr_after is not None else None

        block_objs = [block.obj_ref for block in blocks]
        after_obj = None if after is None else after.obj_ref
        block_objs, after_block_objs = session.api.blocks.children.append(parent.obj_ref, block_objs, after=after_obj)
        parent.obj_ref.has_children = True

        if parent_node.is_root and isinstance(parent, ChildrenMixin):
            parent._children = [] if parent._children is None else parent._children
            # update the parent's children cache
            if after is None:
                parent._children.extend(blocks)
            else:
                insert_idx = next(idx for idx, block in enumerate(parent._children) if block.id == after.id) + 1
                # we also update the blocks after the position we inserted.
                for block, updated_block_obj in zip(parent._children[insert_idx:], after_block_objs, strict=True):
                    block.obj_ref.update(**updated_block_obj.model_dump())
                parent._children[insert_idx:insert_idx] = blocks
                curr_after = blocks[-1]  # update to the last inserted block to continue appending after it

        for block_node, block_obj in zip(parent_node.children, block_objs, strict=True):
            # update the appended blocks with the returned objects from the API. This only works at the top level.
            block_node.block.obj_ref.update(**block_obj.model_dump())
            session.cache[block_node.block.id] = block_node.block
            if block_node.children:  # we need to update explicitly the first level for Table, Columns, etc.
                child_objs = session.api.blocks.children.list(block_node.block.obj_ref)
                for child_node, child_obj in zip(block_node.children, child_objs, strict=True):
                    child_node.block.obj_ref.update(**child_obj.model_dump())
                    session.cache[child_node.block.id] = child_node.block



"""Errors/exceptions of the Ultimate Notion API."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ultimate_notion.schema import Schema


class UltimateNotionError(Exception):
    """Base class for all exceptions in this package."""


class UnsetError(UltimateNotionError):
    """Raised when an unset value is accessed before being initialized by the Notion API."""


class SessionError(UltimateNotionError):
    """Raised when there are issues with the Notion session."""


class UnknownUserError(UltimateNotionError):
    """Raised when the user is unknown."""


class UnknownPageError(UltimateNotionError):
    """Raised when the page is unknown."""


class InvalidAPIUsageError(UltimateNotionError):
    """Raised when the API is used in an invalid way."""


class EmptyDBError(UltimateNotionError):
    """A special exception that tells us that a database is empty during probing."""


class FilterQueryError(ValueError, UltimateNotionError):
    """An exception that is raised when a filter query is invalid."""


class EmptyListError(UltimateNotionError):
    """Custom exception for an empty list in SList."""


class MultipleItemsError(UltimateNotionError):
    """Custom exception for a list with multiple items in SList."""


class SchemaError(UltimateNotionError):
    """Raised when there are issues with the schema of a database."""


class SchemaNotBoundError(SchemaError):
    """Raised when the schema is not bound to a database."""

    def __init__(self, schema: type[Schema]):
        self.schema = schema
        msg = f'Schema {schema.__name__} is not bound to any database'
        super().__init__(msg)


class RollupError(SchemaError):
    """Error if definition of rollup is wrong."""


class RelationError(SchemaError):
    """Error if a Relation cannot be initialised."""


class ReadOnlyPropertyError(SchemaError):
    """Raised when an attempt is made to write to a write-protected property."""


class PropertyError(SchemaError):
    """Raised when there is an issue with a property in the schema."""



"""Provides an object-based Notion API with all endpoints.

This pydantic based API is often referred to as `api` while the low-level
API of the [Notion Client SDK library](https://github.com/ramnes/notion-sdk-py)
is referred to as `raw_api`.
"""

from __future__ import annotations

import logging
from collections.abc import Iterator, Mapping
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, BinaryIO, cast
from uuid import UUID

from pydantic import SerializeAsAny, TypeAdapter

from ultimate_notion.obj_api.blocks import Block, Database, FileBase, Page
from ultimate_notion.obj_api.core import Unset, UnsetType, raise_unset
from ultimate_notion.obj_api.enums import FileUploadMode, FileUploadStatus
from ultimate_notion.obj_api.iterator import EndpointIterator, PropertyItemList
from ultimate_notion.obj_api.objects import (
    Bot,
    Comment,
    CustomEmojiObject,
    DatabaseRef,
    EmojiObject,
    FileObject,
    FileUpload,
    GenericObject,
    ObjectRef,
    PageRef,
    ParentRef,
    RichTextBaseObject,
    User,
    UserRef,
)
from ultimate_notion.obj_api.props import PropertyItem, PropertyValue, Title
from ultimate_notion.obj_api.query import DBQueryBuilder, SearchQueryBuilder
from ultimate_notion.obj_api.schema import Property, RenameProp

if TYPE_CHECKING:
    from notion_client import Client as NCClient
    from notion_client.api_endpoints import BlocksChildrenEndpoint as NCBlocksChildrenEndpoint
    from notion_client.api_endpoints import BlocksEndpoint as NCBlocksEndpoint
    from notion_client.api_endpoints import CommentsEndpoint as NCCommentsEndpoint
    from notion_client.api_endpoints import DatabasesEndpoint as NCDatabasesEndpoint
    from notion_client.api_endpoints import FileUploadsEndpoint as NCFileUploadsEndpoint
    from notion_client.api_endpoints import PagesEndpoint as NCPagesEndpoint
    from notion_client.api_endpoints import PagesPropertiesEndpoint as NCPagesPropertiesEndpoint
    from notion_client.api_endpoints import UsersEndpoint as NCUsersEndpoint

_logger = logging.getLogger(__name__)


class NotionAPI:
    """Object-based Notion API (pydantic) with all endpoints."""

    def __init__(self, client: NCClient):
        self.client = client
        self.blocks = BlocksEndpoint(self)
        self.databases = DatabasesEndpoint(self)
        self.pages = PagesEndpoint(self)
        self.search = SearchEndpoint(self)
        self.users = UsersEndpoint(self)
        self.comments = CommentsEndpoint(self)
        self.uploads = UploadsEndpoint(self)


@dataclass
class Endpoint:
    """Baseclass of the Notion API endpoints."""

    api: NotionAPI


class BlocksEndpoint(Endpoint):
    """Interface to the 'blocks' endpoint of the Notion API"""

    class ChildrenEndpoint(Endpoint):
        """Interface to the API 'blocks/children' endpoint."""

        @property
        def raw_api(self) -> NCBlocksChildrenEndpoint:
            """Return the underlying endpoint in the Notion SDK."""
            return self.api.client.blocks.children

        # https://developers.notion.com/reference/patch-block-children
        def append(
            self, parent: ParentRef | GenericObject | UUID | str, blocks: list[Block], *, after: Block | None = None
        ) -> tuple[list[Block], list[Block]]:
            """Add the given blocks as children of the specified parent.

            The blocks info of the passed blocks will be updated and returned as first part of a tuple.
            The second party of the tuple is an empty list or the updated blocks after the specified block
            if `after` was specified. Use this to update the blocks with the latest version from the server.
            """
            parent_id = ObjectRef.build(parent).id
            children = [block.serialize_for_api() for block in blocks if block is not None]
            _logger.debug(f'Appending {len(children)} blocks to parent with id `{parent_id}`.')

            block_iter = EndpointIterator[Block](endpoint=self.raw_api.append, pagination=self.raw_api.list)
            if after is None:
                appended_blocks = list(block_iter(block_id=parent_id, children=children))
                if len(appended_blocks) != len(blocks):
                    msg = 'Number of appended blocks does not match the number of provided blocks.'
                    raise ValueError(msg)
            else:
                appended_blocks = list(
                    block_iter(block_id=parent_id, children=children, after=str(raise_unset(after.id)))
                )

            # the first len(blocks) of appended_blocks correspond to the blocks we passed, the rest are updated
            # blocks after the specified block, where we append the blocks.
            for block, appended_block in zip(blocks, appended_blocks[: len(blocks)], strict=True):
                block.update(**appended_block.serialize_for_api())

            return blocks, appended_blocks[len(blocks) :]

        # https://developers.notion.com/reference/get-block-children
        def list(self, parent: ParentRef | GenericObject | UUID | str) -> Iterator[Block]:
            """Return all Blocks contained by the specified parent."""
            parent_id = ObjectRef.build(parent).id
            _logger.debug(f'Listing all blocks for parent with id `{parent_id}`.')
            block_iter = EndpointIterator[Block](endpoint=self.raw_api.list)
            return block_iter(block_id=parent_id)

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the `blocks` endpoint for the Notion API."""
        super().__init__(*args, **kwargs)
        self.children = BlocksEndpoint.ChildrenEndpoint(*args, **kwargs)

    @property
    def raw_api(self) -> NCBlocksEndpoint:
        """Return the underlying endpoint in the Notion SDK."""
        return self.api.client.blocks

    # https://developers.notion.com/reference/delete-a-block
    def delete(self, block: Block | UUID | str) -> Block:
        """Delete (archive) the specified Block."""
        block_id = str(ObjectRef.build(block).id)
        _logger.debug(f'Deleting block with id `{block_id}`.')
        data = self.raw_api.delete(block_id)
        return Block.model_validate(data)

    def restore(self, block: Block | UUID | str) -> Block:
        """Restore (unarchive) the specified Block."""
        block_id = str(ObjectRef.build(block).id)
        _logger.debug(f'Restoring block with id `{block_id}`.')
        data = self.raw_api.update(block_id, archived=False)
        return Block.model_validate(data)

    # https://developers.notion.com/reference/retrieve-a-block
    def retrieve(self, block: Block | UUID | str) -> Block:
        """Return the requested Block."""
        block_id = str(ObjectRef.build(block).id)
        _logger.debug(f'Retrieving block with id `{block_id}`.')
        data = self.raw_api.retrieve(block_id)
        return Block.model_validate(data)

    # https://developers.notion.com/reference/update-a-block
    def update(self, block: Block) -> None:
        """Update the block object on the server.

        The block info will be updated to the latest version from the server.
        """
        block_id = cast(UUID, raise_unset(block.id))  # don't get why mypy needs this cast
        _logger.debug(f'Updating block with id `{block_id}`.')
        params = block.serialize_for_api()

        if isinstance(block, FileBase):
            # The Notiopn API does not support setting a new typed FileObject, e.g. `external` or `file`
            # It even must be removed from the params
            dtype = params[block.type].pop('type')
            del params[block.type][dtype]

        # Typing in notion_client sucks, so we cast
        data = cast(dict[str, Any], self.raw_api.update(block_id.hex, **params))
        block.update(**data)


class DatabasesEndpoint(Endpoint):
    """Interface to the 'databases' endpoint of the Notion API."""

    @property
    def raw_api(self) -> NCDatabasesEndpoint:
        """Return the underlying endpoint in the Notion SDK."""
        return self.api.client.databases

    @staticmethod
    def _build_request(
        *,
        parent: SerializeAsAny[ParentRef] | None = None,
        schema: Mapping[str, Property | RenameProp | None] | None = None,
        title: list[RichTextBaseObject] | None = None,
        description: list[RichTextBaseObject] | None = None,
        inline: bool | None = None,
    ) -> dict[str, Any]:
        """Build a request payload from the given items.

        *NOTE* this method does not anticipate what the request will be used for and as
        such does not validate the inputs for any particular requests.
        """
        request: dict[str, Any] = {}

        if parent is not None:
            request['parent'] = parent.serialize_for_api()

        if title is not None:
            request['title'] = [rt_obj.serialize_for_api() for rt_obj in title]

        if description is not None:
            request['description'] = [rt_obj.serialize_for_api() for rt_obj in description]

        if schema is not None:
            request['properties'] = {
                name: value.serialize_for_api() if value is not None else None for name, value in schema.items()
            }

        if inline is not None:
            request['is_inline'] = inline

        return request

    # https://developers.notion.com/reference/create-a-database
    def create(
        self,
        parent: Page,
        schema: Mapping[str, Property],
        *,
        title: list[RichTextBaseObject] | None = None,
        inline: bool = False,
    ) -> Database:
        """Add a database to the given Page parent."""
        parent_ref = PageRef.build(parent)
        _logger.debug(f'Creating new database below page with id `{parent_ref.page_id}`.')
        request = self._build_request(parent=parent_ref, schema=schema, title=title, inline=inline)
        data = self.raw_api.create(**request)
        return Database.model_validate(data)

    # https://developers.notion.com/reference/retrieve-a-database
    def retrieve(self, dbref: Database | str | UUID) -> Database:
        """Return the Database with the given ID."""
        db_id = DatabaseRef.build(dbref).database_id
        _logger.debug(f'Retrieving database with id `{db_id}`.')
        data = self.raw_api.retrieve(str(db_id))
        return Database.model_validate(data)

    def update(
        self,
        db: Database,
        *,
        title: list[RichTextBaseObject] | None = None,
        description: list[RichTextBaseObject] | None = None,
        inline: bool | None = None,
        schema: Mapping[str, Property | RenameProp | None] | None = None,
    ) -> None:
        """Update the Database object on the server.

        The database info will be updated to the latest version from the server.

        API reference: https://developers.notion.com/reference/update-a-database
        """
        _logger.debug(f'Updating info of database with id `{db.id}`.')

        if request := self._build_request(schema=schema, title=title, description=description, inline=inline):
            # https://github.com/ramnes/notion-sdk-py/blob/main/notion_client/api_endpoints.py
            # Typing in notion_client sucks, thus we cast
            data = cast(dict[str, Any], self.raw_api.update(str(db.id), **request))
            db.update(**data)

    def delete(self, db: Database) -> Database:
        """Delete (archive) the specified Database."""
        db_id = DatabaseRef.build(db).database_id
        _logger.debug(f'Deleting database with id `{db_id}`.')
        block_obj = self.api.blocks.delete(str(db_id))
        # block.update(**data) is not possible as the API returns a block, not a database
        db.archived = block_obj.archived  # ToDo: Remove when `archived` is completely deprecated
        db.in_trash = block_obj.in_trash
        return db

    def restore(self, db: Database) -> Database:
        """Restore (unarchive) the specified Database."""
        db_id = DatabaseRef.build(db).database_id
        _logger.debug(f'Restoring database with id `{db_id}`.')
        block_obj = self.api.blocks.restore(str(db_id))
        # block.update(**data) is not possible as the API returns a block, not a database
        db.archived = block_obj.archived  # ToDo: Remove when `archived` is completely deprecated
        db.in_trash = block_obj.in_trash
        return db

    # https://developers.notion.com/reference/post-database-query
    def query(self, db: Database | UUID | str) -> DBQueryBuilder:
        """Initialize a new Query object with the target data class."""
        db_id = DatabaseRef.build(db).database_id
        _logger.debug(f'Initializing query for db with id `{db_id}`.')
        return DBQueryBuilder(endpoint=self.raw_api.query, db_id=str(db_id))


class PagesEndpoint(Endpoint):
    """Interface to the API 'pages' endpoint."""

    class PropertiesEndpoint(Endpoint):
        """Interface to the API 'pages/properties' endpoint."""

        @property
        def raw_api(self) -> NCPagesPropertiesEndpoint:
            """Return the underlying endpoint in the Notion SDK"""
            return self.api.client.pages.properties

        # https://developers.notion.com/reference/retrieve-a-page-property
        def retrieve(self, page: Page | UUID | str, property: PropertyValue | str) -> Iterator[PropertyItem]:  # noqa: A002
            """Return the Property on a specific Page with the given ID"""
            page_id = str(PageRef.build(page).page_id)
            property_id = property.id if isinstance(property, PropertyValue) else property
            _logger.debug(f'Retrieving property with id `{property_id}` from page with id `{page_id}`.')
            prop_iter = EndpointIterator[PropertyItem](
                endpoint=self.raw_api.retrieve,
                model_validate=TypeAdapter(PropertyItemList | PropertyItem).validate_python,
            )
            return prop_iter(page_id=page_id, property_id=property_id)

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the `pages` endpoint for the Notion API"""
        super().__init__(*args, **kwargs)
        self.properties = PagesEndpoint.PropertiesEndpoint(*args, **kwargs)

    @property
    def raw_api(self) -> NCPagesEndpoint:
        """Return the underlying endpoint in the Notion SDK"""
        return self.api.client.pages

    # https://developers.notion.com/reference/post-page
    def create(
        self,
        parent: ParentRef | Page | Database,
        title: Title | None = None,
        properties: dict[str, PropertyValue] | None = None,
        children: list[Block] | None = None,
    ) -> Page:
        """Add a page to the given parent (Page or Database)."""
        if parent is None:
            msg = "'parent' must be provided"
            raise ValueError(msg)

        match parent:
            case Page():
                parent = PageRef.build(parent)
                parent_id = parent.page_id
            case Database():
                parent = DatabaseRef.build(parent)
                parent_id = parent.database_id
            case _:
                msg = f'Unsupported parent of type {type(parent)}'
                raise ValueError(msg)

        request: dict[str, Any] = {'parent': parent.serialize_for_api()}

        # the API requires a properties object, even if empty
        if properties is None:
            properties = {}

        if title is not None:
            properties['title'] = title

        request['properties'] = {
            name: prop.serialize_for_api() if prop is not None else None for name, prop in properties.items()
        }

        if children is not None:
            request['children'] = [child.serialize_for_api() for child in children if child is not None]

        _logger.debug(f'Creating new page below page with id `{parent_id}`.')
        data = self.raw_api.create(**request)
        return Page.model_validate(data)

    def delete(self, page: Page) -> None:
        """Delete (archive) the specified Page."""
        self.set_attr(page, in_trash=True)

    def restore(self, page: Page) -> None:
        """Restore (unarchive) the specified Page."""
        self.set_attr(page, in_trash=False)

    # https://developers.notion.com/reference/retrieve-a-page
    def retrieve(self, page: Page | UUID | str) -> Page:
        """Return the requested Page.

        !!! warning

            This method will only retrieve up to 25 items per property.
            Use `pages.properties.retrieve` to retrieve all items of a specific property.
        """
        page_id = str(PageRef.build(page).page_id)
        _logger.debug(f'Retrieving page with id `{page_id}`.')
        data = self.raw_api.retrieve(page_id)
        return Page.model_validate(data)

    # https://developers.notion.com/reference/patch-page
    def update(self, page: Page, **properties: PropertyValue) -> None:
        """Update the Page object properties on the server."""
        page_id = raise_unset(page.id)
        _logger.debug(f'Updating info on page with id `{page_id}`.')
        props = {name: value.serialize_for_api() if value is not None else None for name, value in properties.items()}
        data = cast(dict[str, Any], self.raw_api.update(page_id.hex, properties=props))
        page.update(**data)

    def set_attr(
        self,
        page: Page,
        *,
        cover: FileObject | UnsetType | None = Unset,
        icon: FileObject | EmojiObject | CustomEmojiObject | UnsetType | None = Unset,
        in_trash: bool | UnsetType = Unset,
    ) -> None:
        """Set specific page attributes (such as cover, icon, etc.) on the server.

        `page` may be any suitable `PageRef` type.

        To remove an attribute, set its value to None.
        """
        page_id = PageRef.build(page).page_id
        props: dict[str, Any] = {}

        if cover is not Unset:
            if cover is None:
                _logger.debug(f'Removing cover from page with id `{page_id}`.')
                props['cover'] = None
            else:
                _logger.debug(f'Setting cover on page with id `{page_id}`.')
                props['cover'] = cover.serialize_for_api()  # type: ignore[union-attr]

        if icon is not Unset:
            if icon is None:
                _logger.debug(f'Removing icon from page with id `{page_id}`.')
                props['icon'] = None
            else:
                _logger.debug(f'Setting icon on page with id `{page_id}`.')
                props['icon'] = icon.serialize_for_api()  # type: ignore[union-attr]

        if in_trash is not Unset:
            if in_trash:
                _logger.debug(f'Deleting page with id `{page_id}`.')
                props['archived'] = True
            else:
                _logger.debug(f'Restoring page with id `{page_id}`.')
                props['archived'] = False

        data = cast(dict[str, Any], self.raw_api.update(page_id.hex, **props))
        page.update(**data)


class SearchEndpoint(Endpoint):
    """Interface to the API 'search' endpoint."""

    # https://developers.notion.com/reference/post-search
    def __call__(self, text: str | None = None) -> SearchQueryBuilder:
        """Perform a search with the optional text"""
        return SearchQueryBuilder(endpoint=self.api.client.search, text=text)


class UsersEndpoint(Endpoint):
    """Interface to the API 'users' endpoint."""

    @property
    def raw_api(self) -> NCUsersEndpoint:
        """Return the underlying endpoint in the Notion SDK"""
        return self.api.client.users

    # https://developers.notion.com/reference/get-users
    def list(self) -> Iterator[User]:
        """Return an iterator for all users in the workspace."""
        _logger.debug('Retrieving all known users.')
        user_iter = EndpointIterator[User](endpoint=self.raw_api.list)
        return user_iter()

    # https://developers.notion.com/reference/get-user
    def retrieve(self, user: User | UUID | str) -> User:
        """Return the User with the given ID."""
        user_id = str(UserRef.build(user).id)
        _logger.debug(f'Retrieving user with id `{user_id}`.')
        data = self.raw_api.retrieve(user_id)
        return User.model_validate(data)

    # https://developers.notion.com/reference/get-self
    def me(self) -> Bot:
        """Return the current bot User."""
        _logger.debug('Retrieving current integration bot')
        data = self.raw_api.me()
        return Bot.model_validate(data)


class CommentsEndpoint(Endpoint):
    """Interface to the API 'comments' endpoint."""

    @property
    def raw_api(self) -> NCCommentsEndpoint:
        """Return the underlying endpoint in the Notion SDK"""
        return self.api.client.comments

    # https://developers.notion.com/reference/create-a-comment
    def create(self, page: Page | UUID | str, rich_text: list[RichTextBaseObject]) -> Comment:
        """Create a comment on the specified Page."""
        page_ref = PageRef.build(page)
        _logger.debug(f'Creating a comment on page with id `{page_ref.page_id}.')
        rich_text_json = [rt.serialize_for_api() for rt in rich_text]
        data = self.raw_api.create(parent=page_ref.serialize_for_api(), rich_text=rich_text_json)
        return Comment.model_validate(data)

    # https://developers.notion.com/reference/create-a-comment
    def append(self, discussion_id: UUID | str, rich_text: list[RichTextBaseObject]) -> Comment:
        """Append a comment to the specified discussion."""
        _logger.debug(f'Appending a comment to discussion with id `{discussion_id}`.')
        rich_text_json = [rt.serialize_for_api() for rt in rich_text]
        data = self.raw_api.create(discussion_id=str(discussion_id), rich_text=rich_text_json)
        return Comment.model_validate(data)

    # https://developers.notion.com/reference/retrieve-a-comment
    def list(self, block: Block | Page | UUID | str) -> Iterator[Comment]:
        """Return all comments on the specified page or block."""
        block_id = str(ObjectRef.build(block).id)
        _logger.debug(f'Listing comments on block with id `{block_id}`.')
        comment_iter = EndpointIterator[Comment](endpoint=self.raw_api.list)
        return comment_iter(block_id=block_id)


class UploadsEndpoint(Endpoint):
    """Interface to the API 'file uploads' endpoint."""

    @property
    def raw_api(self) -> NCFileUploadsEndpoint:
        """Return the underlying endpoint in the Notion SDK"""
        return self.api.client.file_uploads

    # https://developers.notion.com/reference/create-a-file-upload
    def create(
        self,
        name: str | None = None,
        n_parts: int | None = None,
        mode: FileUploadMode = FileUploadMode.SINGLE_PART,
        external_url: str | None = None,
        content_type: str | None = None,
    ) -> FileUpload:
        """Create a file upload."""
        _logger.debug(f'Creating a file upload with mode `{mode}`.')
        kwargs = {
            'filename': name,
            'number_of_parts': n_parts,
            'mode': mode.value,
            'external_url': external_url,
            'content_type': content_type,
        }
        kwargs = {k: v for k, v in kwargs.items() if v is not None}  # Notion API doesn't like nulls
        data = self.raw_api.create(**kwargs)
        return FileUpload.model_validate(data)

    # https://developers.notion.com/reference/send-a-file-upload
    def send(self, file_upload: FileUpload, file: BinaryIO, part: int | None = None) -> None:
        """Send a file upload and update the file_upload object."""
        _logger.debug(f'Sending file upload with id `{file_upload.id}`.')
        file_param = (file_upload.filename, file, file_upload.content_type)  # see issue #127 for reasons
        # Notion API doesn't like nulls, so we eliminate them
        kwargs = {k: v for k, v in (('file', file_param), ('part_number', part)) if v is not None}
        data = cast(dict[str, Any], self.raw_api.send(file_upload_id=str(file_upload.id), **kwargs))
        file_upload.update(**data)

    # https://developers.notion.com/reference/complete-a-file-upload
    def complete(self, file_upload: FileUpload) -> None:
        """Complete the file upload and update the file_upload object."""
        _logger.debug(f'Completing file upload with id `{file_upload.id}`.')
        data = cast(dict[str, Any], self.raw_api.complete(file_upload_id=str(file_upload.id)))
        file_upload.update(**data)

    # https://developers.notion.com/reference/retrieve-a-file-upload
    def retrieve(self, upload_id: UUID | str) -> FileUpload:
        """Return the FileUpload with the given ID."""
        _logger.debug(f'Retrieving file upload with id `{upload_id}`.')
        data = self.raw_api.retrieve(file_upload_id=str(upload_id))
        return FileUpload.model_validate(data)

    # https://developers.notion.com/reference/list-file-uploads
    def list(self, status: FileUploadStatus | None = None, page_size: int = 100) -> Iterator[FileUpload]:
        """Return all file uploads."""
        file_upload_iter = EndpointIterator[FileUpload](endpoint=self.raw_api.list)
        kwargs: dict[str, Any] = {'page_size': page_size}
        if status is not None:
            kwargs['status'] = status.value
            suffix = f' filtering by status {status.value}'
        else:
            suffix = '.'
        _logger.debug(f'Listing all file uploads{suffix}')
        return file_upload_iter(**kwargs)



"""Provides an interactive query builder for Notion databases."""

from __future__ import annotations

import logging
from abc import ABC
from collections.abc import Awaitable, Callable, Iterator, Mapping
from datetime import date, datetime
from typing import Any, Generic, Literal, TypeAlias
from uuid import UUID

from pydantic import ConfigDict, Field, SerializeAsAny, field_validator
from typing_extensions import TypeVar

from ultimate_notion.obj_api.blocks import Database, Page
from ultimate_notion.obj_api.core import GenericObject
from ultimate_notion.obj_api.enums import SortDirection, TimestampKind
from ultimate_notion.obj_api.iterator import MAX_PAGE_SIZE, EndpointIterator

NCEndpointCall: TypeAlias = Callable[..., Any | Awaitable[Any]]  # ToDo: `type` instead of `TypeAlias` in Python 3.12

_logger = logging.getLogger(__name__)


class Condition(GenericObject, ABC):
    """Base class for all conditions in Notion."""


class TextCondition(Condition):
    """Represents text criteria in Notion."""

    equals: str | None = None
    does_not_equal: str | None = None
    contains: str | None = None
    does_not_contain: str | None = None
    starts_with: str | None = None
    ends_with: str | None = None
    is_empty: bool | None = None
    is_not_empty: bool | None = None


class NumberCondition(Condition):
    """Represents number criteria in Notion."""

    equals: float | int | None = None
    does_not_equal: float | int | None = None
    greater_than: float | int | None = None
    less_than: float | int | None = None
    greater_than_or_equal_to: float | int | None = None
    less_than_or_equal_to: float | int | None = None
    is_empty: bool | None = None
    is_not_empty: bool | None = None


class IdCondition(Condition):
    """Represents ID criteria in Notion."""

    equals: float | int | None = None
    does_not_equal: float | int | None = None
    greater_than: float | int | None = None
    less_than: float | int | None = None
    greater_than_or_equal_to: float | int | None = None
    less_than_or_equal_to: float | int | None = None


class CheckboxCondition(Condition):
    """Represents checkbox criteria in Notion."""

    equals: bool | None = None
    does_not_equal: bool | None = None


class SelectCondition(Condition):
    """Represents select criteria in Notion."""

    equals: str | None = None
    does_not_equal: str | None = None
    is_empty: bool | None = None
    is_not_empty: bool | None = None


class MultiSelectCondition(Condition):
    """Represents a multi_select criteria in Notion."""

    contains: str | None = None
    does_not_contain: str | None = None
    is_empty: bool | None = None
    is_not_empty: bool | None = None


class DateCondition(Condition):
    """Represents date criteria in Notion."""

    equals: date | datetime | None = None
    before: date | datetime | None = None
    after: date | datetime | None = None
    on_or_before: date | datetime | None = None
    on_or_after: date | datetime | None = None

    is_empty: bool | None = None
    is_not_empty: bool | None = None

    class EmptyObject(GenericObject): ...

    this_week: EmptyObject | None = None
    past_week: EmptyObject | None = None
    past_month: EmptyObject | None = None
    past_year: EmptyObject | None = None
    next_week: EmptyObject | None = None
    next_month: EmptyObject | None = None
    next_year: EmptyObject | None = None


class PeopleCondition(Condition):
    """Represents people criteria in Notion."""

    contains: UUID | None = None
    does_not_contain: UUID | None = None
    is_empty: bool | None = None
    is_not_empty: bool | None = None


class FilesCondition(Condition):
    """Represents files criteria in Notion."""

    is_empty: bool | None = None
    is_not_empty: bool | None = None


class RelationCondition(Condition):
    """Represents relation criteria in Notion."""

    contains: UUID | None = None
    does_not_contain: UUID | None = None
    is_empty: bool | None = None
    is_not_empty: bool | None = None


class FormulaCondition(Condition):
    """Represents formula criteria in Notion."""

    string: TextCondition | None = None
    checkbox: CheckboxCondition | None = None
    number: NumberCondition | None = None
    date: DateCondition | None = None


class RollupArrayCondition(Condition):
    """Represents a rollup array filter in Notion."""

    rich_text: TextCondition | None = None
    phone_number: TextCondition | None = None
    number: NumberCondition | None = None
    checkbox: CheckboxCondition | None = None
    select: SelectCondition | None = None
    multi_select: MultiSelectCondition | None = None
    date: DateCondition | None = None
    people: PeopleCondition | None = None
    files: FilesCondition | None = None
    relation: RelationCondition | None = None
    formula: FormulaCondition | None = None
    unique_id: IdCondition | None = None


class RollupCondition(Condition):
    """Represents rollup criteria in Notion."""

    any: RollupArrayCondition | None = None
    every: RollupArrayCondition | None = None
    none: RollupArrayCondition | None = None
    date: DateCondition | None = None
    number: NumberCondition | None = None


class QueryFilter(GenericObject):
    """Base class for query filters."""


class PropertyFilter(QueryFilter, RollupArrayCondition):
    """Represents a database property filter in Notion."""

    property: str
    rollup: RollupCondition | None = None


class SearchFilter(QueryFilter):
    """Represents a search property filter in Notion."""

    property: str
    value: str


class TimestampFilter(QueryFilter):
    """Represents a timestamp filter in Notion."""

    timestamp: TimestampKind


class CreatedTimeFilter(TimestampFilter):
    """Represents a created_time filter in Notion."""

    created_time: DateCondition
    timestamp: TimestampKind = TimestampKind.CREATED_TIME


class LastEditedTimeFilter(TimestampFilter):
    """Represents a last_edited_time filter in Notion."""

    last_edited_time: DateCondition
    timestamp: TimestampKind = TimestampKind.LAST_EDITED_TIME


class CompoundFilter(QueryFilter):
    """Represents a compound filter in Notion."""

    model_config = ConfigDict(populate_by_name=True)

    and_: list[SerializeAsAny[QueryFilter]] | None = Field(None, alias='and')
    or_: list[SerializeAsAny[QueryFilter]] | None = Field(None, alias='or')


class DBSort(GenericObject):
    """Sort instruction when querying a database"""

    property: str
    direction: SortDirection


class SearchSort(GenericObject):
    """Sort instruction when searching for pages and databases"""

    timestamp: Literal[TimestampKind.LAST_EDITED_TIME]
    direction: SortDirection


class Query(GenericObject, ABC):
    """Abstract query object in Notion for searching pages/databases and querying databases"""

    filter: SerializeAsAny[QueryFilter] | None = None
    start_cursor: UUID | None = None
    page_size: int = MAX_PAGE_SIZE

    @field_validator('page_size')
    @classmethod
    def valid_page_size(cls, value: int) -> int:
        """Validate that the given page size meets the Notion API requirements"""

        if value <= 0:
            msg = 'page size must be greater than zero'
            raise ValueError(msg)
        if value > MAX_PAGE_SIZE:
            msg = f'page size must be less than or equal to {MAX_PAGE_SIZE}'
            raise ValueError(msg)

        return value


class SearchQuery(Query):
    """Query object in Notion for searching pages & databases"""

    sort: SearchSort | None = None


class DBQuery(Query):
    """Query object in Notion for querying a database"""

    sorts: list[DBSort] | None = None


T = TypeVar('T', bound=Page | Database, default=Page | Database)


class QueryBuilder(Generic[T], ABC):
    """General query builder for the Notion search & database query API"""

    query: Query
    endpoint: NCEndpointCall
    params: dict[str, str]

    def __init__(self, endpoint: NCEndpointCall, query: Query, params: Mapping[str, str | None]):
        self.endpoint = endpoint
        self.query = query
        self.params = {k: v for k, v in params.items() if v is not None}  # API doesn't like "undefined" values

    def execute(self, **nc_params: int | str) -> Iterator[T]:
        """Execute the current query and return an iterator for the results."""
        query = self.query.serialize_for_api()
        query |= self.params | nc_params
        return EndpointIterator[T](self.endpoint)(**query)


class SearchQueryBuilder(QueryBuilder[T]):
    """Search query builder to search for pages and databases

    By default and not changed by `sort`, then the most recently edited results are returned first.

    Notion API: https://developers.notion.com/reference/post-search
    """

    query: SearchQuery

    def __init__(self, endpoint: NCEndpointCall, text: str | None = None):
        super().__init__(endpoint=endpoint, query=SearchQuery(), params={'query': text})

    def execute(self, **nc_params: int | str) -> Iterator[T]:
        match self.filter:
            case SearchFilter(property='object', value='page'):
                _logger.debug(f'Searching for pages with title: {self.params["query"]}')
            case SearchFilter(property='object', value='database'):
                _logger.debug(f'Searching for databases with title: {self.params["query"]}')
            case None:
                _logger.debug(f'Searching for pages and databases with title: {self.params["query"]}')
        return super().execute(**nc_params)

    def filter(self, *, page_only: bool = False, db_only: bool = False) -> SearchQueryBuilder[T]:
        """Filter for pages or databases only"""
        if not (page_only ^ db_only):
            msg = 'Either `page_only` or `db_only` must be true, not both.'
            raise ValueError(msg)
        elif page_only:
            value = 'page'
        else:  # db_only
            value = 'database'

        builder = SearchQueryBuilder[T](self.endpoint, text=self.params.get('query'))
        builder.query.filter = SearchFilter(property='object', value=value)
        return builder

    def sort(self, *, ascending: bool) -> SearchQueryBuilder[T]:
        """Add the given sort elements to the query."""
        direction = SortDirection.ASCENDING if ascending else SortDirection.DESCENDING

        builder = SearchQueryBuilder[T](self.endpoint, text=self.params.get('query'))
        builder.query.sort = SearchSort(timestamp=TimestampKind.LAST_EDITED_TIME, direction=direction)
        return builder


class DBQueryBuilder(QueryBuilder[Page]):
    """Query builder to query a database.

    Notion API: https://developers.notion.com/reference/post-search
    """

    query: DBQuery

    def __init__(self, endpoint: NCEndpointCall, db_id: str):
        super().__init__(endpoint=endpoint, query=DBQuery(), params={'database_id': db_id})

    def execute(self, **nc_params: int | str) -> Iterator[Page]:
        _logger.debug(f'Searching for pages in database with id: {self.params["database_id"]}')
        return super().execute(**nc_params)

    def filter(self, condition: QueryFilter) -> DBQueryBuilder:
        """Add the given filter to the query."""
        builder = DBQueryBuilder(self.endpoint, db_id=self.params['database_id'])
        if self.query.sorts is not None:
            builder.query.sorts = [sort.model_copy(deep=True) for sort in self.query.sorts]
        builder.query.filter = condition
        return builder

    def sort(self, sort_orders: DBSort | list[DBSort]) -> DBQueryBuilder:
        """Add the given sort elements to the query."""
        if isinstance(sort_orders, DBSort):
            sort_orders = [sort_orders]

        builder = DBQueryBuilder(self.endpoint, db_id=self.params['database_id'])
        if self.query.filter is not None:
            builder.query.filter = self.query.filter.model_copy(deep=True)
        builder.query.sorts = sort_orders
        return builder



"""Functionality for working with Notion databases."""

from __future__ import annotations

from typing import Any

from pydantic import ValidationError
from typing_extensions import Self

from ultimate_notion.blocks import ChildrenMixin, DataObject, wrap_icon
from ultimate_notion.core import get_active_session, get_repr
from ultimate_notion.emoji import CustomEmoji, Emoji
from ultimate_notion.errors import ReadOnlyPropertyError, SchemaError
from ultimate_notion.file import AnyFile
from ultimate_notion.obj_api import blocks as obj_blocks
from ultimate_notion.obj_api.core import raise_unset
from ultimate_notion.page import Page
from ultimate_notion.query import Query
from ultimate_notion.rich_text import Text, camel_case
from ultimate_notion.schema import Property, Schema
from ultimate_notion.view import View


class Database(DataObject[obj_blocks.Database], wraps=obj_blocks.Database):
    """A Notion database.

    This object always represents an original database, not a linked database.

    API reference: https://developers.notion.com/docs/working-with-databases
    """

    _schema: type[Schema] | None = None

    def __str__(self) -> str:
        if self.title:
            return str(self.title)
        else:
            return 'Untitled'

    def _repr_html_(self) -> str:  # noqa: PLW3201
        """Called by JupyterLab automatically"""
        return str(self)

    def __repr__(self) -> str:
        return get_repr(self)

    @property
    def url(self) -> str:
        """Return the URL of this database."""
        return raise_unset(self.obj_ref.url)

    @property
    def title(self) -> str | Text | None:
        """Return the title of this database as rich text."""
        # `str` added as return value but always RichText returned, which inherits from str.
        if title := raise_unset(self.obj_ref.title):
            return Text.wrap_obj_ref(title)
        return None

    @title.setter
    def title(self, text: str | Text | None) -> None:
        """Set the title of this database"""
        if text is None:
            text = ''
        if not isinstance(text, Text):
            text = Text.from_plain_text(text)
        session = get_active_session()
        session.api.databases.update(self.obj_ref, title=text.obj_ref)

    @property
    def description(self) -> Text | None:
        """Return the description of this database as rich text."""
        if desc := raise_unset(self.obj_ref.description):
            return Text.wrap_obj_ref(desc)
        return None

    @description.setter
    def description(self, text: str | Text | None) -> None:
        """Set the description of this database."""
        if text is None:
            text = ''
        if not isinstance(text, Text):
            text = Text.from_plain_text(text)
        session = get_active_session()
        session.api.databases.update(self.obj_ref, description=text.obj_ref)

    @property
    def icon(self) -> AnyFile | Emoji | CustomEmoji | None:
        """Return the icon of this database as file or emoji."""
        if (icon := self.obj_ref.icon) is None:
            return None
        else:
            return wrap_icon(icon)

    @property
    def cover(self) -> AnyFile | None:
        """Return the cover of this database as file."""
        cover = self.obj_ref.cover
        return AnyFile.wrap_obj_ref(cover) if cover is not None else None

    @property
    def is_wiki(self) -> bool:
        """Return whether the database is a wiki."""
        # ToDo: Implement using the verification property
        raise NotImplementedError

    @property
    def is_db(self) -> bool:
        """Return whether the object is a database."""
        return True

    def _reflect_schema(self, obj_ref: obj_blocks.Database) -> type[Schema]:
        """Reflection about the database schema."""
        title = str(self)
        cls_name = f'{camel_case(title)}Schema'
        attrs = {'_props': [Property.wrap_obj_ref(v) for v in obj_ref.properties.values()]}
        schema: type[Schema] = type(cls_name, (Schema,), attrs, db_title=title)
        schema._bind_db(self)
        return schema

    def _set_schema(self, schema: type[Schema], *, during_init: bool) -> None:
        """Set a custom schema in order to change the Python variables names."""
        self.schema.assert_consistency_with(schema, during_init=during_init)
        schema._bind_db(self)
        self._schema = schema

    @property
    def schema(self) -> type[Schema]:
        """Schema of the database."""
        if not self._schema:
            self._schema = self._reflect_schema(self.obj_ref)
        return self._schema

    @schema.setter
    def schema(self, schema: type[Schema]) -> None:
        """Set a custom schema in order to change the Python variables names."""
        self._set_schema(schema, during_init=False)

    @property
    def is_inline(self) -> bool:
        """Return whether the database is inline."""
        return self.obj_ref.is_inline

    @is_inline.setter
    def is_inline(self, inline: bool) -> None:
        """Set whether the database is inline."""
        if self.is_inline != inline:
            session = get_active_session()
            session.api.databases.update(self.obj_ref, inline=inline)
            self.obj_ref.is_inline = inline

    def delete(self) -> Self:
        """Delete this database."""
        if not self.is_deleted:
            session = get_active_session()
            session.api.databases.delete(self.obj_ref)
            self._delete_me_from_parent()
        return self

    def restore(self) -> Self:
        """Restore this database."""
        if self.is_deleted:
            session = get_active_session()
            session.api.databases.restore(self.obj_ref)
            if isinstance(self.parent, ChildrenMixin):
                self.parent._children = None  # this forces a new retrieval of children next time
        return self

    def reload(self, *, rebind_schema: bool = True) -> Self:
        """Reload this database.

        If `rebind_schema` is `True`, the schema will be rebound to the current database.
        Otherwise, the schema will set to the reflected schema of the current database.
        """
        old_schema = self.schema
        session = get_active_session()
        self.obj_ref = session.api.databases.retrieve(self.id)
        self._schema = None  # reset schema to reflect the current database
        self.schema._set_obj_refs()
        if rebind_schema:
            self.schema = old_schema
        return self

    @property
    def query(self) -> Query:
        """Return a Query object to build and execute a database query."""
        return Query(database=self)

    def get_all_pages(self) -> View:
        """Retrieve all pages and return a view."""
        return self.query.execute()

    def __len__(self) -> int:
        """Return the number of pages in this database."""
        return len(self.get_all_pages())

    @property
    def is_empty(self) -> bool:
        """Return whether the database is empty."""
        return len(self) == 0

    def __bool__(self) -> bool:
        """Overwrite default behaviour."""
        msg = 'Use .is_empty instead of bool(db) to check if a database is empty.'
        raise RuntimeError(msg)

    def create_page(self, **kwargs: Any) -> Page:
        """Create a page with properties according to the schema within the corresponding database."""
        attr_to_name = {prop.attr_name: prop.name for prop in self.schema.get_props()}
        if not set(kwargs).issubset(set(attr_to_name)):
            add_kwargs = set(kwargs) - set(attr_to_name)
            msg = f'Attributes {", ".join(add_kwargs)} not defined for properties in schema {self.__class__.__name__}'
            raise SchemaError(msg)
        if ro_props := set(kwargs) & {prop.attr_name for prop in self.schema.get_ro_props()}:
            msg = f'Read-only properties {", ".join(ro_props)} cannot be set'
            raise ReadOnlyPropertyError(msg)

        schema_kwargs = {attr_to_name[attr]: value for attr, value in kwargs.items()}
        validator = self.schema.to_pydantic_model(with_ro_props=False)
        try:
            schema = validator(**schema_kwargs)
        except ValidationError as e:
            msg = f'Invalid keyword arguments or read-only properties are overwritten:\n{e}'
            raise SchemaError(msg) from e

        session = get_active_session()
        page_obj = session.api.pages.create(parent=self.obj_ref, properties=schema.to_dict())
        page = Page.wrap_obj_ref(page_obj)
        session.cache[page.id] = page
        return page

    def to_markdown(self) -> str:
        """Return the reference to this database as Markdown."""
        return self._to_markdown()

    def _to_markdown(self) -> str:
        """Return the reference to this database as Markdown."""
        return f'[**🗄️ {self.title}**]({self.block_url})\n'



"""Comments and discussions for pages, blocks, and databases."""

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, overload
from uuid import UUID

from typing_extensions import Self

from ultimate_notion.core import NotionEntity, get_active_session
from ultimate_notion.errors import InvalidAPIUsageError
from ultimate_notion.obj_api import objects as objs
from ultimate_notion.rich_text import Text
from ultimate_notion.user import User

if TYPE_CHECKING:
    from ultimate_notion.blocks import DataObject


class Comment(NotionEntity[objs.Comment], wraps=objs.Comment):
    """A comment on a page, block, or database behaving like a normal string."""

    @property
    def text(self) -> Text | None:
        """The text of the comment."""
        if (rt := self.obj_ref.rich_text) is None:
            return None
        return Text.wrap_obj_ref(rt)

    @property
    def user(self) -> User:
        """The user who created the comment."""
        return self.created_by

    @property
    def discussion_id(self) -> UUID:
        """The ID of the discussion thread that this comment belongs to."""
        return self.obj_ref.discussion_id

    def __str__(self) -> str:
        if self.text is None:
            return ''
        return str(self.text)

    def __repr__(self) -> str:
        return f'Comment("{self.user}", "{self!s}")'


class Discussion(Sequence[Comment]):
    """A list of comments, i.e. a discussion thread."""

    def __init__(self, comments: Sequence[Comment], *, parent: DataObject) -> None:
        self._comments = list(comments)
        self._parent = parent

    @overload
    def __getitem__(self, idx: int, /) -> Comment: ...

    @overload
    def __getitem__(self, idx: slice, /) -> Sequence[Comment]: ...

    def __getitem__(self, idx: int | slice, /) -> Comment | Sequence[Comment]:
        return self._comments[idx]

    def __len__(self) -> int:
        return len(self._comments)

    def __repr__(self) -> str:
        return f'Discussion({self._comments})'

    def __str__(self) -> str:
        return '\n'.join(f'{self.user}: {self.text}' for self in self._comments)

    def append(self, text: str) -> Self:
        """Add a comment to the discussion.

        !!! note

            This functionality requires that your integration was granted *insert* comment capabilities.

        """
        if self._parent.is_deleted:
            msg = 'Cannot add a comment to a deleted parent.'
            raise RuntimeError(msg)

        if not self._parent.in_notion:
            msg = 'Cannot add a comment to a parent that is not in Notion.'
            raise RuntimeError(msg)

        text = Text(text)
        session = get_active_session()

        if self._parent.is_page:
            comment = session.api.comments.create(page=self._parent.id, rich_text=text.obj_ref)
        elif self:  # parent is a block that already has comments
            comment = session.api.comments.append(
                rich_text=text.obj_ref, discussion_id=self._comments[-1].discussion_id
            )
        else:
            msg = 'Cannot create a new discussion thread for a block, only append to an existing discussion.'
            raise InvalidAPIUsageError(msg)

        self._comments.append(Comment.wrap_obj_ref(comment))
        return self



"""Ultimate Notion provides a pythonic, high-level API for Notion.

Notion-API: https://developers.notion.com/reference/intro
"""

from __future__ import annotations

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version('ultimate-notion')
except PackageNotFoundError:  # pragma: no cover
    __version__ = 'unknown'
finally:
    del version, PackageNotFoundError

from ultimate_notion.blocks import (
    PDF,
    Audio,
    Block,
    Bookmark,
    Breadcrumb,
    BulletedItem,
    Callout,
    Code,
    Column,
    Columns,
    Divider,
    Embed,
    Equation,
    File,
    Heading1,
    Heading2,
    Heading3,
    Image,
    LinkPreview,
    LinkToPage,
    NumberedItem,
    Paragraph,
    Quote,
    SyncedBlock,
    Table,
    TableOfContents,
    TableRow,
    ToDoItem,
    ToggleItem,
    Video,
)
from ultimate_notion.core import Workspace, WorkspaceType, get_active_session
from ultimate_notion.database import Database
from ultimate_notion.emoji import Emoji
from ultimate_notion.file import AnyFile, ExternalFile, NotionFile, url
from ultimate_notion.obj_api.enums import (
    AggFunc,
    BGColor,
    CodeLang,
    Color,
    FileUploadStatus,
    NumberFormat,
    OptionGroupType,
    VState,
)
from ultimate_notion.option import Option, OptionGroup, OptionNS
from ultimate_notion.page import Page
from ultimate_notion.query import Condition, prop
from ultimate_notion.rich_text import join, math, mention, text
from ultimate_notion.schema import Property, PropType, Schema, SelfRef
from ultimate_notion.session import Session
from ultimate_notion.user import User
from ultimate_notion.utils import DateTimeOrRange, SList

__all__ = [
    'PDF',
    'AggFunc',
    'AnyFile',  # for type hinting only
    'Audio',
    'BGColor',
    'Block',  # for type hinting only
    'Bookmark',
    'Breadcrumb',
    'BulletedItem',
    'Callout',
    'Code',
    'CodeLang',
    'Color',
    'Column',
    'Columns',
    'Condition',  # for type hinting only
    'Database',
    'DateTimeOrRange',  # for type hinting only
    'Divider',
    'Embed',
    'Emoji',
    'Equation',
    'ExternalFile',
    'File',
    'FileUploadStatus',
    'Heading1',
    'Heading2',
    'Heading3',
    'Image',
    'LinkPreview',
    'LinkToPage',
    'NotionFile',
    'NumberFormat',
    'NumberedItem',
    'Option',
    'OptionGroup',
    'OptionGroupType',
    'OptionNS',
    'Page',  # for type hinting only
    'Paragraph',
    'PropType',
    'Property',  # for type hinting only
    'Quote',
    'SList',
    'Schema',
    'SelfRef',
    'Session',
    'SyncedBlock',
    'Table',
    'TableOfContents',
    'TableRow',
    'ToDoItem',
    'ToggleItem',
    'User',
    'VState',
    'Video',
    'Workspace',
    'WorkspaceType',  # for type hinting only
    '__version__',
    'get_active_session',
    'join',
    'math',
    'mention',
    'prop',
    'text',
    'url',
]



"""
Page property values especially for pages within databases.
The names of the properties reflect the name in the Notion UI.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, ClassVar, cast

import pendulum as pnd
from typing_extensions import Self, TypeVar

import ultimate_notion.obj_api.props as obj_props
from ultimate_notion import rich_text as rt
from ultimate_notion.core import Wrapper, get_active_session, get_repr, raise_unset
from ultimate_notion.file import AnyFile
from ultimate_notion.obj_api.enums import FormulaType, RollupType, VState
from ultimate_notion.obj_api.objects import DateRange, DateTimeOrRange
from ultimate_notion.option import Option
from ultimate_notion.user import User

if TYPE_CHECKING:
    from ultimate_notion.page import Page

PV_co = TypeVar('PV_co', bound=obj_props.PropertyValue, default=obj_props.PropertyValue, covariant=True)


class PropertyValue(Wrapper[PV_co], ABC, wraps=obj_props.PropertyValue):  # noqa: PLW1641
    """Base class for Notion property values.

    Used to map high-level objects to low-level Notion-API objects
    """

    readonly: bool = False  # value of property can not be set by us
    _type_value_map: ClassVar[dict[str, type[PropertyValue]]] = {}

    def __init_subclass__(cls, wraps: type[PV_co], **kwargs: Any):
        super().__init_subclass__(wraps=wraps, **kwargs)
        # When this is called, the model is not yet constructed, thus no direct field access with .type.
        type_name = wraps.model_fields['type'].get_default()
        cls._type_value_map[type_name] = cls

    @property
    def _obj_api_type(self) -> type[obj_props.PropertyValue]:
        return self._obj_api_map_inv[self.__class__]

    def __init__(self, values: Any | Sequence[Any]):
        if isinstance(values, Sequence) and not isinstance(values, str | bytes):
            values = [value.obj_ref if isinstance(value, Wrapper) else value for value in values]
        else:
            values = values.obj_ref if isinstance(values, Wrapper) else values

        super().__init__(values)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, PropertyValue):
            return NotImplemented
        other_obj_ref = cast(PV_co, other.obj_ref)
        return (self.obj_ref.type == other_obj_ref.type) and (self.obj_ref.value == other_obj_ref.value)

    @property
    @abstractmethod
    def value(self) -> Any:
        """Return the actual Python value object of this property."""

    @property
    def id(self) -> str:
        return raise_unset(self.obj_ref.id)

    def __repr__(self) -> str:
        return get_repr(self)

    def __str__(self) -> str:
        if isinstance(self.value, rt.Text):
            return str(self.value) if self.value else ''
        elif isinstance(self.value, list):
            # workaround as `str` on lists calls `repr` instead of `str`
            return ', '.join(str(val) for val in self.value)
        else:
            return str(self.value) if self.value else ''


#############################
# Writeable Property Values #
#############################


class Title(PropertyValue[obj_props.Title], wraps=obj_props.Title):
    """Title property value."""

    def __init__(self, text: str):
        super().__init__(rt.Text(text).obj_ref)

    @property
    def value(self) -> rt.Text:
        return rt.Text.wrap_obj_ref(self.obj_ref.title)


class Text(PropertyValue[obj_props.RichText], wraps=obj_props.RichText):
    """Rich text property value."""

    def __init__(self, text: str):
        super().__init__(rt.Text(text).obj_ref)

    @property
    def value(self) -> rt.Text:
        return rt.Text.wrap_obj_ref(self.obj_ref.rich_text)


class Number(PropertyValue[obj_props.Number], wraps=obj_props.Number):
    """Number property value."""

    @property
    def value(self) -> int | float | None:
        return self.obj_ref.number


class Checkbox(PropertyValue[obj_props.Checkbox], wraps=obj_props.Checkbox):
    """Simple checkbox type; represented as a boolean."""

    @property
    def value(self) -> bool | None:
        return self.obj_ref.checkbox


class Date(PropertyValue[obj_props.Date], wraps=obj_props.Date):
    """Date(-time) property value."""

    def __init__(self, dt_spec: str | DateTimeOrRange):
        self.obj_ref = obj_props.Date.build(dt_spec)

    @property
    def value(self) -> DateTimeOrRange | None:
        date = self.obj_ref.date
        if date is None:
            return None
        else:
            return date.to_pendulum()


class Status(PropertyValue[obj_props.Status], wraps=obj_props.Status):
    """Status property value."""

    def __init__(self, option: str | Option):
        if isinstance(option, str):
            option = Option(option)

        super().__init__(option)

    @property
    def value(self) -> Option | None:
        if self.obj_ref.status:
            return Option.wrap_obj_ref(self.obj_ref.status)
        else:
            return None


class Select(PropertyValue[obj_props.Select], wraps=obj_props.Select):
    """Single select property value."""

    def __init__(self, option: str | Option):
        if isinstance(option, str):
            option = Option(option)

        super().__init__(option)

    @property
    def value(self) -> Option | None:
        if self.obj_ref.select:
            return Option.wrap_obj_ref(self.obj_ref.select)
        else:
            return None


class MultiSelect(PropertyValue[obj_props.MultiSelect], wraps=obj_props.MultiSelect):
    """Multi-select property value."""

    def __init__(self, options: str | Option | Sequence[str | Option]):
        if not isinstance(options, Sequence | str) or isinstance(options, str):
            options = [options]
        options = [Option(option) if isinstance(option, str) else option for option in options]
        super().__init__(options)

    @property
    def value(self) -> list[Option] | None:
        if self.obj_ref.multi_select:
            return [Option.wrap_obj_ref(option) for option in self.obj_ref.multi_select]
        else:
            return None


class Person(PropertyValue[obj_props.People], wraps=obj_props.People):
    """Person/People property value."""

    def __init__(self, users: User | Sequence[User]):
        if not isinstance(users, Sequence):
            users = [users]
        super().__init__(users)

    @property
    def value(self) -> list[User]:
        session = get_active_session()
        return [session.get_user(user.id, raise_on_unknown=False) for user in self.obj_ref.people]


class URL(PropertyValue[obj_props.URL], wraps=obj_props.URL):
    """URL property value."""

    @property
    def value(self) -> str | None:
        return self.obj_ref.url


class Email(PropertyValue[obj_props.Email], wraps=obj_props.Email):
    """Email property value."""

    @property
    def value(self) -> str | None:
        return self.obj_ref.email


class Phone(PropertyValue[obj_props.PhoneNumber], wraps=obj_props.PhoneNumber):
    """Phone property value."""

    @property
    def value(self) -> str | None:
        return self.obj_ref.phone_number


class Files(PropertyValue[obj_props.Files], wraps=obj_props.Files):
    """Files property value."""

    def __init__(self, files: AnyFile | Sequence[AnyFile]):
        if not isinstance(files, Sequence):
            files = [files]

        super().__init__(files)

    @property
    def value(self) -> list[AnyFile]:
        return [AnyFile.wrap_obj_ref(file) for file in self.obj_ref.files]


class Relations(PropertyValue[obj_props.Relation], wraps=obj_props.Relation):
    """Relation property values."""

    def __init__(self, pages: Page | Sequence[Page]):
        if not isinstance(pages, Sequence):
            pages = [pages]
        super().__init__(pages)

    @property
    def value(self) -> list[Page]:
        session = get_active_session()
        return [session.get_page(ref_obj.id) for ref_obj in self.obj_ref.relation]


#############################
# Read-Only Property Values #
#############################


class Formula(PropertyValue[obj_props.Formula], wraps=obj_props.Formula):
    """Formula property value."""

    readonly = True

    @property
    def value(self) -> str | float | int | DateTimeOrRange | None:
        """Return the result value of the formula."""
        # returns all values of subclasses of `FormulaResult`
        if self.obj_ref.formula is None:
            return None
        match value := self.obj_ref.formula.value:
            case DateRange():
                return value.to_pendulum() if value else None
            case str() | int() | float():
                return value
            case _:
                return None

    @property
    def value_type(self) -> FormulaType | None:
        """Return the type of the formula result."""
        return FormulaType(self.obj_ref.formula.type) if self.obj_ref.formula else None


class Rollup(PropertyValue[obj_props.Rollup], wraps=obj_props.Rollup):
    """Rollup property value."""

    readonly = True

    @property
    def value(self) -> float | int | DateTimeOrRange | list[Any] | None:
        """Return the result value of the rollup."""
        match rollup_type := self.obj_ref.rollup:
            case None:
                return None
            case obj_props.RollupArray():
                return [PropertyValue.wrap_obj_ref(prop).value for prop in rollup_type.array]
            case obj_props.RollupNumber():
                return rollup_type.number
            case obj_props.RollupDate():
                return rollup_type.date.to_pendulum() if rollup_type.date is not None else None
            case _:
                msg = f'Unknown rollup value type: {type(rollup_type)}'
                raise ValueError(msg)

    @property
    def value_type(self) -> RollupType | None:
        """Return the type of the rollup result."""
        return RollupType(self.obj_ref.rollup.type) if self.obj_ref.rollup else None


class CreatedTime(PropertyValue[obj_props.CreatedTime], wraps=obj_props.CreatedTime):
    """Created-time property value."""

    readonly = True

    @property
    def value(self) -> pnd.DateTime:
        return pnd.instance(self.obj_ref.created_time)


class CreatedBy(PropertyValue[obj_props.CreatedBy], wraps=obj_props.CreatedBy):
    """Created-by property value."""

    readonly = True

    @property
    def value(self) -> User:
        session = get_active_session()
        return session.get_user(self.obj_ref.created_by.id, raise_on_unknown=False)


class LastEditedTime(PropertyValue[obj_props.LastEditedTime], wraps=obj_props.LastEditedTime):
    """Last-edited-time property value."""

    readonly = True

    @property
    def value(self) -> pnd.DateTime:
        return pnd.instance(self.obj_ref.last_edited_time)


class LastEditedBy(PropertyValue[obj_props.LastEditedBy], wraps=obj_props.LastEditedBy):
    """Last-edited-by property value."""

    readonly = True

    @property
    def value(self) -> User:
        session = get_active_session()
        return session.get_user(self.obj_ref.last_edited_by.id, raise_on_unknown=False)


class ID(PropertyValue[obj_props.UniqueID], wraps=obj_props.UniqueID):
    """Unique ID property value."""

    readonly = True

    @property
    def number(self) -> int:
        return self.obj_ref.unique_id.number

    @property
    def prefix(self) -> str | None:
        return self.obj_ref.unique_id.prefix

    @property
    def value(self) -> str:
        return f'{self.prefix}-{self.number}'


class Verification(PropertyValue[obj_props.Verification], wraps=obj_props.Verification):
    """Verification property value of pages in wiki databases."""

    # ToDo: Write a unit test for this!
    readonly = True

    @property
    def value(self) -> Self:
        return self

    @property
    def state(self) -> VState:
        return self.obj_ref.verification.state

    @property
    def verified_by(self) -> User | None:
        if self.obj_ref.verification.verified_by is None:
            return None
        else:
            session = get_active_session()
            return session.get_user(self.obj_ref.verification.verified_by.id, raise_on_unknown=False)

    @property
    def date(self) -> pnd.DateTime | None:
        if self.obj_ref.verification.date is None:
            return None
        else:
            return pnd.instance(self.obj_ref.verification.date)


class Button(PropertyValue[obj_props.Button], wraps=obj_props.Button):
    """Button property value.

    This is a read-only property that represents a button in a database.
    """

    readonly = True

    @property
    def value(self) -> None:
        return None

    def __str__(self) -> str:
        return f'Button(id={self.id})'

    def __repr__(self) -> str:
        return get_repr(self, desc=self.id)



"""Functionality for general Notion objects like texts, files, options, etc."""

from __future__ import annotations

from ultimate_notion.core import Wrapper, get_repr
from ultimate_notion.errors import UnsetError
from ultimate_notion.obj_api import objects as objs
from ultimate_notion.obj_api.core import Unset, UnsetType, raise_unset
from ultimate_notion.obj_api.enums import Color, OptionGroupType
from ultimate_notion.rich_text import Text


class Option(Wrapper[objs.SelectOption], wraps=objs.SelectOption):
    """Option for select & multi-select property."""

    def __init__(self, name: str, *, color: Color | str | UnsetType = Unset) -> None:
        if isinstance(color, str):
            color = Color(color)
        super().__init__(name, color=color)

    def __repr__(self) -> str:
        return get_repr(self)

    def __str__(self) -> str:
        return self.name

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Option):
            res = self.obj_ref == other.obj_ref
        elif isinstance(other, str):
            res = self.name == other
        elif other is None:
            res = False
        else:
            msg = f'Cannot compare Option with types {type(other)}'
            raise RuntimeError(msg)
        return res

    def __hash__(self) -> int:
        return super().__hash__()

    @property
    def id(self) -> str:
        """ID of the option."""
        return raise_unset(self.obj_ref.id)

    @property
    def name(self) -> str:
        """Name of the option."""
        return self.obj_ref.name

    @property
    def color(self) -> Color:
        """Color of the option."""
        try:
            return raise_unset(self.obj_ref.color)
        except UnsetError:  # i.e. unset value
            return Color.DEFAULT

    @property
    def description(self) -> str:
        """Description of the option."""
        if desc := self.obj_ref.description:
            return Text.wrap_obj_ref(desc)
        else:
            return ''


class OptionNSType(type):
    """Metaclass to implement `len` for type `OptionNS` itself, not an instance of it."""

    # ToDo: When mypy is smart enough to understand metaclasses, we can remove the `type: ignore` comments.

    def __len__(cls: type[OptionNS]) -> int:  # type: ignore[misc]
        return len(cls.to_list())


class OptionNS(metaclass=OptionNSType):
    """Option namespace to simplify working with (Multi-)Select options."""

    @classmethod
    def to_list(cls) -> list[Option]:
        """Convert the enum to a list as needed by the (Multi)Select property types."""
        return [
            getattr(cls, var) for var in cls.__dict__ if not var.startswith('__') and not callable(getattr(cls, var))
        ]


class OptionGroup(Wrapper[objs.SelectGroup], wraps=objs.SelectGroup):
    """Group of options for status property."""

    _options: list[Option]  # holds all possible options of this group

    def __init__(self, group_type: OptionGroupType | str, options: list[Option]) -> None:
        group_type = OptionGroupType(group_type)
        # There are three option groups and the colors and names seem to be fixed. No way to change it in the UX.
        match group_type:
            case OptionGroupType.TO_DO:
                name = 'To-do'
                color = Color.GRAY
            case OptionGroupType.IN_PROGRESS:
                name = 'In progress'
                color = Color.BLUE
            case OptionGroupType.COMPLETE:
                name = 'Complete'
                color = Color.GREEN

        self._options = options
        # Note that we use option names (not IDs, which may not be present) to identify options in the group.
        # OptionGroups cannot be created with the Notion API and thus we only use it for convenience
        # during local development and testing and when checking schemas it will be ignored.
        option_ids = []
        for option in options:
            try:
                option_ids.append(option.id)
            except UnsetError:
                option_ids.append(option.name)
        super().__init__(name=name, color=color, option_ids=option_ids)

    @classmethod
    def wrap_obj_ref(cls, obj_ref: objs.SelectGroup, /, *, options: list[Option] | None = None) -> OptionGroup:
        """Convienence constructor for the group of options."""
        obj = super().wrap_obj_ref(obj_ref)
        obj._options = [] if options is None else options
        return obj

    @property
    def name(self) -> str:
        """Name of the option group."""
        return self.obj_ref.name

    @property
    def options(self) -> list[Option]:
        """Options within this option group."""
        return self._options

    def __repr__(self) -> str:
        return get_repr(self)

    def __str__(self) -> str:
        return self.name

    def __eq__(self, other: object) -> bool:
        if isinstance(other, OptionGroup):
            return self.obj_ref == other.obj_ref
        return False

    def __hash__(self) -> int:
        return hash(self.obj_ref)


def check_for_updates(old: list[Option], new: list[Option]) -> dict[str, list[str]]:
    """Check if two lists of options contain updates.

    Returns which attributes have changed for each option. This is mainly used to check if options have changed
    when updating a select or multi-select property as this is not supported by the Notion API.
    """
    old_by_name = {opt.name: opt for opt in old}
    new_by_name = {opt.name: opt for opt in new}
    common_names = set(old_by_name) & set(new_by_name)

    updates: dict[str, list[str]] = {}
    for name in common_names:
        old_opt = old_by_name[name]
        new_opt = new_by_name[name]

        for attr in ('name', 'color', 'description'):
            if getattr(old_opt, attr) != getattr(new_opt, attr):
                updates.setdefault(name, []).append(attr)
    return updates
